#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""utility classes for the Builder
"""
import glob
import os
import re
from pathlib import Path

import numpy
from wx.lib.agw.aui.aui_constants import *
import wx.lib.statbmp
from wx.lib.agw.aui.aui_utilities import IndentPressedBitmap, ChopText, TakeScreenShot
import sys
import wx
import wx.lib.agw.aui as aui
from wx.lib import platebtn

import psychopy
from psychopy import logging
from . import pavlovia_ui
from . import icons
from .themes import ThemeMixin, IconCache
from psychopy.localization import _translate
from psychopy.tools.stringtools import prettyname
from psychopy.tools.apptools import SortTerm


class FileDropTarget(wx.FileDropTarget):
    """On Mac simply setting a handler for the EVT_DROP_FILES isn't enough.
    Need this too.
    """

    def __init__(self, targetFrame):
        wx.FileDropTarget.__init__(self)
        self.target = targetFrame
        self.app = targetFrame.app

    def OnDropFiles(self, x, y, filenames):
        logging.debug(
            'PsychoPyBuilder: received dropped files: %s' % filenames)
        for fname in filenames:
            if isinstance(self.target, psychopy.app.coder.CoderFrame) and wx.GetKeyState(wx.WXK_ALT):
                # If holding ALT and on coder, insert filename into current coder doc
                if self.app.coder:
                    if self.app.coder.currentDoc:
                        self.app.coder.currentDoc.AddText(fname)
            if isinstance(self.target, psychopy.app.runner.RunnerFrame):
                # If on Runner, load file to run
                self.app.showRunner()
                self.app.runner.addTask(fileName=fname)
            elif fname.lower().endswith('.psyexp'):
                # If they dragged on a .psyexp file, open it in in Builder
                self.app.showBuilder()
                self.app.builder.fileOpen(filename=fname)
            else:
                # If they dragged on any other file, try to open it in Coder (if it's not text, this will give error)
                self.app.showCoder()
                self.app.coder.fileOpen(filename=fname)
        return True


class WindowFrozen():
    """
    Equivalent to wxWindowUpdateLocker.

    Usage::

        with WindowFrozen(wxControl):
            update multiple things
        # will automatically thaw here

    """

    def __init__(self, ctrl):
        self.ctrl = ctrl

    def __enter__(self):  # started the with... statement
        # Freeze should not be called if platform is win32.
        if sys.platform == 'win32':
            return self.ctrl

        # check it hasn't been deleted
        #
        # Don't use StrictVersion() here, as `wx` doesn't follow the required
        # numbering scheme.
        if self.ctrl is not None and wx.__version__[:3] <= '2.8':
            self.ctrl.Freeze()
        return self.ctrl

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Thaw should not be called if platform is win32.
        if sys.platform == 'win32':
            return
        # check it hasn't been deleted
        if self.ctrl is not None and self.ctrl.IsFrozen():
            self.ctrl.Thaw()


def getSystemFonts(encoding='system', fixedWidthOnly=False):
    """Get a list of installed system fonts.

    Parameters
    ----------
    encoding : str
        Get fonts with matching encodings.
    fixedWidthOnly : bool
        Return on fixed width fonts.

    Returns
    -------
    list
        List of font facenames.

    """
    fontEnum = wx.FontEnumerator()

    encoding = "FONTENCODING_" + encoding.upper()
    if hasattr(wx, encoding):
        encoding = getattr(wx, encoding)

    return fontEnum.GetFacenames(encoding, fixedWidthOnly=fixedWidthOnly)


class PsychopyToolbar(wx.ToolBar, ThemeMixin):
    """Toolbar for the Builder/Coder Frame"""
    def __init__(self, frame):
        wx.ToolBar.__init__(self, frame)
        self.frame = frame
        self.app = self.frame.app
        self._needMakeTools = True
        # Configure toolbar appearance
        self.SetWindowStyle(wx.TB_HORIZONTAL | wx.NO_BORDER | wx.TB_FLAT | wx.TB_NODIVIDER)
        # self.SetBackgroundColour(ThemeMixin.appColors['frame_bg'])
        # Set icon size (16 for win/linux small mode, 32 for everything else
        self.iconSize = 32  # mac: 16 either doesn't work, or looks bad
        self.SetToolBitmapSize((self.iconSize, self.iconSize))
        # OS-dependent tool-tips
        ctrlKey = 'Ctrl+'
        if sys.platform == 'darwin':
            ctrlKey = 'Cmd+'
        # keys are the keyboard keys, not the keys of the dict
        self.keys = {k: self.frame.app.keys[k].replace('Ctrl+', ctrlKey)
                for k in self.frame.app.keys}
        self.keys['none'] = ''
        # self.makeTools()  # will be done when theme is applied
        self.buttons = {}
        # Finished setup. Make it happen

    def makeTools(self):
        frame = self.frame
        # Create tools
        cl = frame.__class__.__name__
        pavButtons = pavlovia_ui.toolbar.PavloviaButtons(
                frame, toolbar=self, tbSize=self.iconSize)
        if frame.__class__.__name__ == 'BuilderFrame':
            self.addPsychopyTool(
                    name='filenew',
                    label=_translate('New'),
                    shortcut='new',
                    tooltip=_translate("Create new experiment file"),
                    func=self.frame.app.newBuilderFrame)  # New
            self.addPsychopyTool(
                    name='fileopen',
                    label=_translate('Open'),
                    shortcut='open',
                    tooltip=_translate("Open an existing experiment file"),
                    func=self.frame.fileOpen)  # Open
            self.frame.bldrBtnSave = self.addPsychopyTool(
                        name='filesave',
                        label=_translate('Save'),
                        shortcut='save',
                        tooltip=_translate("Save current experiment file"),
                        func=self.frame.fileSave)  # Save
            self.addPsychopyTool(
                    name='filesaveas',
                    label=_translate('Save As...'),
                    shortcut='saveAs',
                    tooltip=_translate("Save current experiment file as..."),
                    func=self.frame.fileSaveAs)  # SaveAs
            self.frame.bldrBtnUndo = self.addPsychopyTool(
                        name='undo',
                        label=_translate('Undo'),
                        shortcut='undo',
                        tooltip=_translate("Undo last action"),
                        func=self.frame.undo)  # Undo
            self.frame.bldrBtnRedo = self.addPsychopyTool(
                        name='redo',
                        label=_translate('Redo'),
                        shortcut='redo',
                        tooltip=_translate("Redo last action"),
                        func=self.frame.redo)  # Redo
            self.AddSeparator()  # Separator
            self.addPsychopyTool(
                    name='monitors',
                    label=_translate('Monitor Center'),
                    shortcut='none',
                    tooltip=_translate("Monitor settings and calibration"),
                    func=self.frame.app.openMonitorCenter)  # Monitor Center
            self.addPsychopyTool(
                    name='cogwindow',
                    label=_translate('Experiment Settings'),
                    shortcut='none',
                    tooltip=_translate("Edit experiment settings"),
                    func=self.frame.setExperimentSettings)  # Settings
            self.AddSeparator()
            self.addPsychopyTool(
                    name='compile_py',
                    label=_translate('Compile Python Script'),
                    shortcut='compileScript',
                    tooltip=_translate("Compile to Python script"),
                    func=self.frame.compileScript)  # Compile
            self.addPsychopyTool(
                    name='compile_js',
                    label=_translate('Compile JS Script'),
                    shortcut='compileScript',
                    tooltip=_translate("Compile to JS script"),
                    func=self.frame.fileExport)  # Compile
            self.frame.bldrBtnRunner = self.addPsychopyTool(
                    name='runner',
                    label=_translate('Runner'),
                    shortcut='runnerScript',
                    tooltip=_translate("Send experiment to Runner"),
                    func=self.frame.runFile)  # Run
            self.frame.bldrBtnRun = self.addPsychopyTool(
                    name='run',
                    label=_translate('Run'),
                    shortcut='runScript',
                    tooltip=_translate("Run experiment"),
                    func=self.frame.runFile)  # Run
            self.AddSeparator()  # Separator
            pavButtons.addPavloviaTools()
        elif frame.__class__.__name__ == 'CoderFrame':
            self.addPsychopyTool('filenew', _translate('New'), 'new',
                                 _translate("Create new experiment file"),
                                 self.frame.fileNew)  # New
            self.addPsychopyTool('fileopen', _translate('Open'), 'open',
                                 _translate("Open an existing experiment file"),
                                 self.frame.fileOpen)  # Open
            self.frame.cdrBtnSave = \
                self.addPsychopyTool('filesave', _translate('Save'), 'save',
                                     _translate("Save current experiment file"),
                                     self.frame.fileSave)  # Save
            self.addPsychopyTool('filesaveas', _translate('Save As...'), 'saveAs',
                                 _translate("Save current experiment file as..."),
                                 self.frame.fileSaveAs)  # SaveAs
            self.frame.cdrBtnUndo = \
                self.addPsychopyTool('undo', _translate('Undo'), 'undo',
                                     _translate("Undo last action"),
                                     self.frame.undo)  # Undo
            self.frame.cdrBtnRedo = \
                self.addPsychopyTool('redo', _translate('Redo'), 'redo',
                                     _translate("Redo last action"),
                                     self.frame.redo)  # Redo
            self.AddSeparator()  # Separator
            self.addPsychopyTool('monitors', _translate('Monitor Center'), 'none',
                                 _translate("Monitor settings and calibration"),
                                 self.frame.app.openMonitorCenter)
            self.addPsychopyTool('color', _translate('Color Picker'), 'none',
                                 _translate("Color Picker -> clipboard"),
                                 self.frame.app.colorPicker)
            self.AddSeparator()
            self.frame.cdrBtnRunner = self.addPsychopyTool(
                    'runner', _translate('Runner'), 'runnerScript',
                    _translate("Send experiment to Runner"),
                    self.frame.runFile)
            self.frame.cdrBtnRun = self.addPsychopyTool(
                    'run', _translate('Run'), 'runScript',
                    _translate("Run experiment"),
                    self.frame.runFile)
            self.AddSeparator()
            pavButtons.addPavloviaTools(
                buttons=['pavloviaSync', 'pavloviaSearch', 'pavloviaUser'])
        frame.btnHandles.update(pavButtons.btnHandles)
        self.Realize()
        # Disable compile buttons until an experiment is present
        if 'compile_py' in self.buttons:
            self.EnableTool(self.buttons['compile_py'].GetId(), Path(self.frame.filename).is_file())
        if 'compile_js' in self.buttons:
            self.EnableTool(self.buttons['compile_js'].GetId(), Path(self.frame.filename).is_file())

    def addPsychopyTool(self, name, label, shortcut, tooltip, func,
                        emblem=None):
        if not name.endswith('.png'):
            filename = name + '.png'
        else:
            filename = name
        self.buttons[name] = self.app.iconCache.makeBitmapButton(parent=self, filename=filename,
                                                   name=label,
                                                   label=("%s [%s]" % (
                                                       label,
                                                       self.keys[shortcut])),
                                                   emblem=emblem, toolbar=self,
                                                   tip=tooltip,
                                                   size=self.iconSize)
        # Bind function
        self.Bind(wx.EVT_TOOL, func, self.buttons[name])
        return self.buttons[name]


class PsychopyPlateBtn(platebtn.PlateButton, ThemeMixin):
    def __init__(self, parent, id=wx.ID_ANY, label='', bmp=None,
                 pos=wx.DefaultPosition, size=wx.DefaultSize,
                 style=1, name=wx.ButtonNameStr):
        platebtn.PlateButton.__init__(self, parent, id, label, bmp, pos, size, style, name)
        self.parent = parent
        self.__InitColors()
        self._applyAppTheme()

    def _applyAppTheme(self):
        cs = ThemeMixin.appColors
        self.__InitColors()
        self.SetBackgroundColour(wx.Colour(self.parent.GetBackgroundColour()))
        self.SetPressColor(cs['txtbutton_bg_hover'])
        self.SetLabelColor(cs['text'],
                           cs['txtbutton_fg_hover'])

    def __InitColors(self):
        cs = ThemeMixin.appColors
        """Initialize the default colors"""
        colors = dict(default=True,
                      hlight=cs['txtbutton_bg_hover'],
                      press=cs['txtbutton_bg_hover'],
                      htxt=cs['text'])
        return colors


class ButtonArray(wx.Window):

    class ArrayBtn(wx.Window):
        def __init__(self, parent, label=""):
            wx.Window.__init__(self, parent)
            self.parent = parent
            # Setup sizer
            self.sizer = wx.BoxSizer(wx.HORIZONTAL)
            self.SetSizer(self.sizer)
            # Create button
            self.button = wx.Button(self, label=label, style=wx.BORDER_NONE)
            self.sizer.Add(self.button, border=4, flag=wx.LEFT | wx.EXPAND)
            # Create remove btn
            self.removeBtn = wx.Button(self, label="×", size=(24, -1))
            self.sizer.Add(self.removeBtn, border=4, flag=wx.RIGHT | wx.EXPAND)
            # Bind remove btn to remove function
            self.removeBtn.Bind(wx.EVT_BUTTON, self.remove)
            # Bind button to button function
            self.button.Bind(wx.EVT_BUTTON, self.onClick)

            self.SetBackgroundColour(self.parent.GetBackgroundColour())
            self.Layout()

        def remove(self, evt=None):
            self.parent.removeItem(self)

        def onClick(self, evt=None):
            evt = wx.CommandEvent(wx.EVT_BUTTON.typeId)
            evt.SetEventObject(self)
            wx.PostEvent(self.parent, evt)

    def __init__(self, parent, orient=wx.HORIZONTAL,
                 items=(),
                 options=None,
                 itemAlias=_translate("item")):
        # Create self
        wx.Window.__init__(self, parent)
        self.SetBackgroundColour(parent.GetBackgroundColour())
        self.parent = parent
        self.itemAlias = itemAlias
        self.options = options
        # Create sizer
        self.sizer = wx.WrapSizer(orient=orient)
        self.SetSizer(self.sizer)
        # Create add button
        self.addBtn = wx.Button(self, size=(24, 24), label="+", style=wx.BORDER_NONE)
        self.addBtn.Bind(wx.EVT_BUTTON, self.newItem)
        self.sizer.Add(self.addBtn, border=3, flag=wx.EXPAND | wx.ALL)
        # Add items
        self.items = items
        # Layout
        self.Layout()

    def _applyAppTheme(self, target=None):
        for child in self.sizer.Children:
            if hasattr(child.Window, "_applyAppTheme"):
                child.Window._applyAppTheme()

    @property
    def items(self):
        items = {}
        for child in self.sizer.Children:
            if not child.Window == self.addBtn:
                items[child.Window.button.Label] = child.Window
        return items

    @items.setter
    def items(self, value):
        if isinstance(value, str):
            value = [value]
        if value is None or value is numpy.nan:
            value = []
        assert isinstance(value, (list, tuple))

        value.reverse()

        self.clear()
        for item in value:
            self.addItem(item)

    def newItem(self, evt=None):
        msg = _translate("Add {}...").format(self.itemAlias)
        if self.options is None:
            _dlg = wx.TextEntryDialog(self.parent, message=msg)
        else:
            _dlg = wx.SingleChoiceDialog(self.parent, msg, "Input Text", choices=self.options)

        if _dlg.ShowModal() != wx.ID_OK:
            return
        if self.options is None:
            self.addItem(_dlg.GetValue())
        else:
            self.addItem(_dlg.GetStringSelection())

    def addItem(self, item):
        if not isinstance(item, wx.Window):
            item = self.ArrayBtn(self, label=item)
        self.sizer.Insert(0, item, border=3, flag=wx.EXPAND | wx.TOP | wx.BOTTOM | wx.RIGHT)
        self.Layout()
        # Raise event
        evt = wx.ListEvent(wx.EVT_LIST_INSERT_ITEM.typeId)
        evt.SetEventObject(self)
        wx.PostEvent(self, evt)

    def removeItem(self, item):
        items = self.items.copy()
        # Get value from key if needed
        if item in items:
            item = items[item]
        # Delete object and item in dict
        if item in list(items.values()):
            i = self.sizer.Children.index(self.sizer.GetItem(item))
            self.sizer.Remove(i)
            item.Hide()
        self.Layout()
        # Raise event
        evt = wx.ListEvent(wx.EVT_LIST_DELETE_ITEM.typeId)
        evt.SetEventObject(self)
        wx.PostEvent(self, evt)

    def clear(self):
        # Raise event
        evt = wx.ListEvent(wx.EVT_LIST_DELETE_ALL_ITEMS.typeId)
        evt.SetEventObject(self)
        wx.PostEvent(self, evt)
        # Delete all items
        for item in self.items:
            self.removeItem(item)

    def Enable(self, enable=True):
        for child in self.Children:
            child.Enable(enable)

    def Disable(self):
        self.Enable(False)

    def GetValue(self):
        return list(self.items)


class SortCtrl(wx.Window):
    class SortItem(wx.Window):
        def __init__(self, parent,
                     item,
                     showSelect=False, selected=True,
                     showFlip=False
                     ):
            # Create self
            wx.Window.__init__(self, parent, style=wx.BORDER_NONE)
            self.SetBackgroundColour("white")
            self.parent = parent
            # Make sure we've been given a SortTerm
            assert isinstance(item, SortTerm)
            self.item = item
            # Setup sizer
            self.sizer = wx.BoxSizer(wx.HORIZONTAL)
            self.SetSizer(self.sizer)
            # Add tickbox (if select)
            self.selectCtrl = wx.CheckBox(self)
            self.selectCtrl.Bind(wx.EVT_CHECKBOX, self.onSelect)
            self.selectCtrl.SetValue(selected)
            self.selectCtrl.Show(showSelect)
            self.sizer.Add(self.selectCtrl, border=6, flag=wx.ALL | wx.EXPAND)
            # Add label
            self.labelObj = wx.StaticText(self, label=self.item.label)
            self.sizer.Add(self.labelObj, proportion=1, border=6, flag=wx.ALL | wx.EXPAND)
            # Add flip button
            self.flipBtn = wx.Button(self, size=(16, 8), label="⇵", style=wx.BORDER_NONE)
            self.flipBtn.SetBackgroundColour(self.GetBackgroundColour())
            self.flipBtn.Bind(wx.EVT_BUTTON, self.flip)
            self.flipBtn.Show(showFlip)
            self.sizer.Add(self.flipBtn, border=6, flag=wx.ALL | wx.EXPAND)
            # Add ctrls sizer
            self.ctrlsSizer = wx.BoxSizer(wx.VERTICAL)
            self.sizer.Add(self.ctrlsSizer, border=6, flag=wx.ALL | wx.EXPAND)
            # Add up button
            self.upBtn = wx.Button(self, size=(16, 8), label="▲", style=wx.BORDER_NONE)
            self.upBtn.SetBackgroundColour(self.GetBackgroundColour())
            self.upBtn.Bind(wx.EVT_BUTTON, self.moveUp)
            self.ctrlsSizer.Add(self.upBtn, border=0, flag=wx.ALL | wx.EXPAND)
            # Add stretch spacer inbetween
            self.ctrlsSizer.AddStretchSpacer(1)
            # Add up button
            self.downBtn = wx.Button(self, size=(16, 8), label="▼", style=wx.BORDER_NONE)
            self.downBtn.SetBackgroundColour(self.GetBackgroundColour())
            self.downBtn.Bind(wx.EVT_BUTTON, self.moveDown)
            self.ctrlsSizer.Add(self.downBtn, border=0, flag=wx.ALL | wx.EXPAND)
            # Do initial select
            self.onSelect()

        @property
        def label(self):
            return self.item.label

        @property
        def value(self):
            return self.item.value

        def flip(self, evt=None):
            # Flip state
            self.item.ascending = not self.item.ascending
            # Change label
            self.labelObj.SetLabel(self.label)

        def moveUp(self, evt=None):
            # Get own index
            i = self.parent.items.index(self)
            # Insert popped self before previous position
            self.parent.items.insert(max(i-1, 0), self.parent.items.pop(i))
            # Layout
            self.parent.Layout()

        def moveDown(self, evt=None):
            # Get own index
            i = self.parent.items.index(self)
            # Insert popped self before previous position
            self.parent.items.insert(min(i+1, len(self.parent.items)), self.parent.items.pop(i))
            # Layout
            self.parent.Layout()

        @property
        def selected(self):
            return self.selectCtrl.GetValue()

        def onSelect(self, evt=None):
            self.Enable(self.selected)

        def Enable(self, enable=True):
            self.labelObj.Enable(enable)

        def Disable(self):
            self.Enable(False)

    def __init__(self, parent,
                 items,
                 showSelect=False, selected=True,
                 showFlip=False,
                 orient=wx.VERTICAL):
        wx.Window.__init__(self, parent)
        # Make sure we've been given an array
        if not isinstance(items, (list, tuple)):
            items = [items]
        # Setup sizer
        self.sizer = wx.BoxSizer(orient)
        self.SetSizer(self.sizer)
        # If given a bool for select, apply it to all items
        if isinstance(selected, bool):
            selected = [selected] * len(items)
        assert isinstance(selected, (list, tuple)) and len(selected) == len(items)
        # Setup items
        self.items = []
        for i, item in enumerate(items):
            self.items.append(self.SortItem(self,
                                            item=item,
                                            showSelect=showSelect, selected=selected[i],
                                            showFlip=showFlip))
            self.sizer.Add(self.items[i], border=6, flag=wx.ALL | wx.EXPAND)
        # Layout
        self.Layout()

    def GetValue(self):
        items = []
        for item in self.items:
            if item.selected:
                items.append(item.item)
        return items

    def Layout(self):
        # Get order of items in reverse
        items = self.items.copy()
        items.reverse()
        # Remove each item
        for item in items:
            self.sizer.Remove(self.sizer.Children.index(self.sizer.GetItem(item)))
        # Add items back in oder
        for i, item in enumerate(items):
            self.sizer.Prepend(item, border=6, flag=wx.ALL | wx.EXPAND)
            item.upBtn.Show(i != len(items)-1)
            item.downBtn.Show(i != 0)
        # Disable appropriate buttons
        # Do base layout
        wx.Window.Layout(self)


class ImageCtrl(wx.lib.statbmp.GenStaticBitmap):
    def __init__(self, parent, bitmap, size=(128, 128)):
        wx.lib.statbmp.GenStaticBitmap.__init__(self, parent, ID=wx.ID_ANY, bitmap=wx.Bitmap(), size=size)
        self.parent = parent
        self.iconCache = IconCache()
        # Set bitmap
        self.SetBitmap(bitmap)
        # Setup sizer
        self.sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.AddStretchSpacer(1)
        self.SetSizer(self.sizer)
        # Add edit button
        self.editBtn = wx.Button(self, size=(24, 24), label=chr(int("270E", 16)))
        self.editBtn.Bind(wx.EVT_BUTTON, self.LoadBitmap)
        self.sizer.Add(self.editBtn, border=6, flag=wx.ALIGN_BOTTOM | wx.ALL)

    def LoadBitmap(self, evt=None):
        # Open file dlg
        _dlg = wx.FileDialog(self.parent, message=_translate("Select image..."))
        if _dlg.ShowModal() != wx.ID_OK:
            return
        # Get value
        path = str(Path(_dlg.GetPath()))
        self.SetBitmap(path)
        # Post event
        evt = wx.FileDirPickerEvent(wx.EVT_FILEPICKER_CHANGED.typeId, self, -1, path)
        evt.SetEventObject(self)
        wx.PostEvent(self, evt)

    def SetBitmap(self, bitmap):
        # Get from file if needed
        if not isinstance(bitmap, wx.Bitmap):
            self.path = bitmap
            bitmap = wx.Bitmap(bitmap)
        # Sub in blank bitmaps
        if not bitmap.IsOk():
            bitmap = self.iconCache.getBitmap(name="user_none", size=128)
        # Store full size bitmap
        self._fullBitmap = bitmap
        # Resize bitmap
        buffer = bitmap.ConvertToImage()
        buffer = buffer.Scale(*self.Size, quality=wx.IMAGE_QUALITY_HIGH)
        scaledBitmap = wx.Bitmap(buffer)
        # Set image
        wx.lib.statbmp.GenStaticBitmap.SetBitmap(self, scaledBitmap)

    @property
    def path(self):
        """
        If current bitmap is from a file, returns the filepath. Otherwise, returns None.
        """
        if hasattr(self, "_path"):
            return self._path

    @path.setter
    def path(self, value):
        self._path = value

    def GetBitmapFull(self):
        return self._fullBitmap

    @property
    def BitmapFull(self):
        return self.GetBitmapFull()

    def Enable(self, enable=True):
        wx.StaticBitmap.Enable(self, enable)
        self.editBtn.Enable(enable)

    def Disable(self):
        self.Enable(False)


class PsychopyScrollbar(wx.ScrollBar):
    def __init__(self, parent, ori=wx.VERTICAL):
        wx.ScrollBar.__init__(self)
        if ori == wx.HORIZONTAL:
            style = wx.SB_HORIZONTAL
        else:
            style = wx.SB_VERTICAL
        self.Create(parent, style=style)
        self.ori = ori
        self.parent = parent
        self.Bind(wx.EVT_SCROLL, self.DoScroll)
        self.Resize()

    def DoScroll(self, event):
        if self.ori == wx.HORIZONTAL:
            w = event.GetPosition()
            h = self.parent.GetScrollPos(wx.VERTICAL)
        elif self.ori == wx.VERTICAL:
            w = self.parent.GetScrollPos(wx.HORIZONTAL)
            h = event.GetPosition()
        else:
            return
        self.parent.Scroll(w, h)
        self.Resize()

    def Resize(self):
        sz = self.parent.GetSize()
        vsz = self.parent.GetVirtualSize()
        start = self.parent.GetViewStart()
        if self.ori == wx.HORIZONTAL:
            sz = (sz.GetWidth(), 20)
            vsz = vsz.GetWidth()
        elif self.ori == wx.VERTICAL:
            sz = (20, sz.GetHeight())
            vsz = vsz.GetHeight()
        self.SetDimensions(start[0], start[1], sz[0], sz[1])
        self.SetScrollbar(
            position=self.GetScrollPos(self.ori),
            thumbSize=10,
            range=1,
            pageSize=vsz
        )


class FileCtrl(wx.TextCtrl):
    def __init__(self, parent, dlgtype="file", value=""):
        wx.TextCtrl.__init__(self, parent, value=value, size=(-1, 24))
        # Store type
        self.dlgtype = dlgtype
        # Setup sizer
        self.sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(self.sizer)
        self.sizer.AddStretchSpacer(1)
        # Add button
        self.fileBtn = wx.Button(self, size=(16, 16), style=wx.BORDER_NONE)
        self.fileBtn.SetBackgroundColour(self.GetBackgroundColour())
        self.fileBtn.SetBitmap(IconCache().getBitmap(name="folder", size=16))
        self.sizer.Add(self.fileBtn, border=4, flag=wx.ALL)
        # Bind browse function
        self.fileBtn.Bind(wx.EVT_BUTTON, self.browse)

    def browse(self, evt=None):
        file = Path(self.GetValue())
        # Sub in a / for blank paths to force the better folder navigator
        if file == Path():
            file = Path("/")
        # Open file or dir dlg
        if self.dlgtype == "dir":
            dlg = wx.DirDialog(self, message=_translate("Specify folder..."), defaultPath=str(file))
        else:
            dlg = wx.FileDialog(self, message=_translate("Specify file..."), defaultDir=str(file))
        if dlg.ShowModal() != wx.ID_OK:
            return
        # Get data from dlg
        file = Path(dlg.GetPath())
        # Set data
        self.SetValue(str(file))

    def SetValue(self, value):
        # Replace backslashes with forward slashes
        value = value.replace("\\", "/")
        # Do base value setting
        wx.TextCtrl.SetValue(self, value)
        # Post event
        evt = wx.FileDirPickerEvent(wx.EVT_FILEPICKER_CHANGED.typeId, self, -1, value)
        evt.SetEventObject(self)
        wx.PostEvent(self, evt)

    def Enable(self, enable=True):
        wx.TextCtrl.Enable(self, enable)
        self.fileBtn.Enable(enable)
        self.fileBtn.SetBackgroundColour(self.GetBackgroundColour())

    def Disable(self):
        self.Enable(False)

    def Show(self, show=True):
        wx.TextCtrl.Show(self, show)
        self.fileBtn.Show(show)

    def Hide(self):
        self.Show(False)


def updateDemosMenu(frame, menu, folder, ext):
    """Update Demos menu as needed."""
    def _makeButton(parent, menu, demo):
        # Skip if demo name starts with _
        if demo.name.startswith("_"):
            return
        # Create menu button
        item = menu.Append(wx.ID_ANY, demo.name)
        # Store in window's demos list
        parent.demos.update({item.Id: demo})
        # Link button to demo opening function
        parent.Bind(wx.EVT_MENU, parent.demoLoad, item)

    def _makeFolder(parent, menu, folder, ext):
        # Skip if underscore in folder name
        if folder.name.startswith("_"):
            return
        # Create and append menu for this folder
        submenu = wx.Menu()
        menu.AppendSubMenu(submenu, folder.name)
        # Get folder contents
        folderContents = glob.glob(str(folder / '*'))
        for subfolder in sorted(folderContents):
            subfolder = Path(subfolder)
            # Make menu/button for each:
            # subfolder according to whether it contains a psyexp, or...
            # subfile according to whether it matches the ext
            if subfolder.is_dir():
                subContents = glob.glob(str(subfolder / '*'))
                if any(file.endswith(".psyexp") and not file.startswith("_") for file in subContents):
                    _makeButton(parent, submenu, subfolder)
                else:
                    _makeFolder(parent, submenu, subfolder, ext)
            elif subfolder.suffix == ext and not subfolder.name.startswith("_"):
                _makeButton(parent, submenu, subfolder)

    # Make blank dict to store demo details in
    frame.demos = {}
    if not folder:  # if there is no unpacked demos folder then just return
        return

    # Get root folders
    rootGlob = glob.glob(str(Path(folder) / '*'))
    for fdr in rootGlob:
        fdr = Path(fdr)
        # Make menus/buttons recursively for each folder according to whether it contains a psyexp
        if fdr.is_dir():
            folderContents = glob.glob(str(fdr / '*'))
            if any(file.endswith(".psyexp") for file in folderContents):
                _makeButton(frame, menu, fdr)
            else:
                _makeFolder(frame, menu, fdr, ext)

class FrameSwitcher(wx.Menu):
    """Menu for switching between different frames"""
    def __init__(self, parent):
        wx.Menu.__init__(self)
        self.parent = parent
        self.app = parent.app
        self.itemFrames = {}
        # Listen for window switch
        self.next = self.Append(wx.ID_MDI_WINDOW_NEXT,
                                _translate("&Next Window\t%s") % self.app.keys['cycleWindows'],
                                _translate("&Next Window\t%s") % self.app.keys['cycleWindows'])
        self.Bind(wx.EVT_MENU, self.nextWindow, self.next)
        self.AppendSeparator()
        # Add creator options
        self.minItemSpec = [
            {'label': "Builder", 'class': psychopy.app.builder.BuilderFrame, 'method': self.app.showBuilder},
            {'label': "Coder", 'class': psychopy.app.coder.CoderFrame, 'method': self.app.showCoder},
            {'label': "Runner", 'class': psychopy.app.runner.RunnerFrame, 'method': self.app.showRunner},
        ]
        for spec in self.minItemSpec:
            if not isinstance(self.Window, spec['class']):
                item = self.Append(
                    wx.ID_ANY, spec['label'], spec['label']
                )
                self.Bind(wx.EVT_MENU, spec['method'], item)
        self.AppendSeparator()
        self.updateFrames()

    @property
    def frames(self):
        return self.parent.app.getAllFrames()

    def updateFrames(self):
        """Set items according to which windows are open"""
        self.next.Enable(len(self.frames) > 1)
        # Make new items if needed
        for frame in self.frames:
            if frame not in self.itemFrames:
                if frame.filename:
                    label = type(frame).__name__.replace("Frame", "") + ": " + os.path.basename(frame.filename)
                else:
                    label = type(frame).__name__.replace("Frame", "")
                self.itemFrames[frame] = self.AppendRadioItem(wx.ID_ANY, label, label)
                self.Bind(wx.EVT_MENU, self.showFrame, self.itemFrames[frame])
        # Edit items to match frames
        for frame in self.itemFrames:
            item = self.itemFrames[frame]
            if not item:
                continue
            if frame not in self.frames:
                # Disable unused items
                item.Enable(False)
            else:
                # Rename item
                if frame.filename:
                    self.itemFrames[frame].SetItemLabel(
                        type(frame).__name__.replace("Frame", "") + ": " + os.path.basename(frame.filename)
                    )
                else:
                    self.itemFrames[frame].SetItemLabel(
                        type(frame).__name__.replace("Frame", "") + ": None"
                    )
            item.Check(frame == self.Window)
        self.itemFrames = {key: self.itemFrames[key] for key in self.itemFrames if self.itemFrames[key] is not None}

    def showFrame(self, event=None):
        itemFrames = event.EventObject.itemFrames
        frame = [key for key in itemFrames if itemFrames[key].Id == event.Id][0]
        frame.Show(True)
        frame.Raise()
        self.parent.app.SetTopWindow(frame)
        self.updateFrames()

    def nextWindow(self, event=None):
        """Cycle through list of open windows"""
        current = event.EventObject.Window
        i = self.frames.index(current)
        while self.frames[i] == current:
            i -= 1
        self.frames[i].Raise()
        self.frames[i].Show()
        self.updateFrames()


def sanitize(inStr):
    """
    Process a string to remove any sensitive information, i.e. OAUTH keys
    """
    # Key-value pairs of patterns with what to replace them with
    patterns = {
        "https\:\/\/oauth2\:[\d\w]{64}@gitlab\.pavlovia\.org\/.*\.git": "[[OAUTH key hidden]]" # Remove any oauth keys
    }
    # Replace each pattern
    for pattern, repl in patterns.items():
        inStr = re.sub(pattern, repl, inStr)

    return inStr


class HoverMixin:
    """
    Mixin providing methods to handle hover on/off events for a wx.Window based class.
    """
    IsHovered = False

    def SetupHover(self):
        """
        Helper method to setup hovering for this object
        """
        # Bind both hover on and hover off events to the OnHover method
        self.Bind(wx.EVT_ENTER_WINDOW, self.OnHover)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.OnHover)

    def OnHover(self, evt=None):
        """
        Method to handle hover events for buttons. To use, bind both `wx.EVT_ENTER_WINDOW` and `wx.EVT_LEAVE_WINDOW` events to this method.
        """
        if evt is None:
            # If calling without event, style according to last IsHovered measurement
            if self.IsHovered:
                self.SetForegroundColour(self.ForegroundColourHover)
                self.SetBackgroundColour(self.BackgroundColourHover)
            else:
                self.SetForegroundColour(self.ForegroundColourNoHover)
                self.SetBackgroundColour(self.BackgroundColourNoHover)
        elif evt.EventType == wx.EVT_ENTER_WINDOW.typeId:
            # If hovered over currently, use hover colours
            self.SetForegroundColour(self.ForegroundColourHover)
            self.SetBackgroundColour(self.BackgroundColourHover)
            # and mark as hovered
            self.IsHovered = True
        else:
            # Otherwise, use regular colours
            self.SetForegroundColour(self.ForegroundColourNoHover)
            self.SetBackgroundColour(self.BackgroundColourNoHover)
            # and mark as unhovered
            self.IsHovered = False

    @property
    def ForegroundColourNoHover(self):
        if hasattr(self, "_ForegroundColourNoHover"):
            return self._ForegroundColourNoHover
        return ThemeMixin.appColors['text']

    @ForegroundColourNoHover.setter
    def ForegroundColourNoHover(self, value):
        self._ForegroundColourNoHover = value

    @property
    def BackgroundColourNoHover(self):
        if hasattr(self, "_BackgroundColourNoHover"):
            return self._BackgroundColourNoHover
        return ThemeMixin.appColors['frame_bg']

    @BackgroundColourNoHover.setter
    def BackgroundColourNoHover(self, value):
        self._BackgroundColourNoHover = value

    @property
    def ForegroundColourHover(self):
        if hasattr(self, "_ForegroundColourHover"):
            return self._ForegroundColourHover
        return ThemeMixin.appColors['txtbutton_fg_hover']

    @ForegroundColourHover.setter
    def ForegroundColourHover(self, value):
        self._ForegroundColourHover = value

    @property
    def BackgroundColourHover(self):
        if hasattr(self, "_BackgroundColourHover"):
            return self._BackgroundColourHover
        return ThemeMixin.appColors['txtbutton_bg_hover']

    @BackgroundColourHover.setter
    def BackgroundColourHover(self, value):
        self._BackgroundColourHover = value


class ToggleButton(wx.ToggleButton, HoverMixin):
    """
    Extends wx.ToggleButton to give methods for handling color changes relating to hover events and value setting.
    """
    @property
    def BackgroundColourNoHover(self):
        if self.GetValue():
            # Return a darker color if selected
            return ThemeMixin.appColors['docker_bg']
        else:
            # Return the default color otherwise
            return HoverMixin.BackgroundColourNoHover.fget(self)


class ToggleButtonArray(wx.Window, ThemeMixin):

    def __init__(self, parent, labels=None, values=None, multi=False, ori=wx.HORIZONTAL):
        wx.Window.__init__(self, parent)
        self.parent = parent
        self.multi = multi
        # Setup sizer
        self.sizer = wx.BoxSizer(ori)
        self.SetSizer(self.sizer)
        # Alias values and labels
        if labels is None:
            labels = values
        if values is None:
            values = labels
        if values is None and labels is None:
            values = labels = []
        # Make buttons
        self.buttons = {}
        for i, val in enumerate(values):
            self.buttons[val] = ToggleButton(self, style=wx.BORDER_NONE)
            self.buttons[val].SetupHover()
            self.buttons[val].SetLabelText(labels[i])
            self.buttons[val].Bind(wx.EVT_TOGGLEBUTTON, self.processToggle)
            self.sizer.Add(self.buttons[val], border=6, proportion=1, flag=wx.ALL | wx.EXPAND)

    def processToggle(self, evt):
        obj = evt.GetEventObject()
        if self.multi:
            # Toggle self
            self.SetValue(self.GetValue())
        else:
            # Selectself and deselect other buttons
            for key, btn in self.buttons.items():
                if btn == obj:
                    self.SetValue(key)

    def SetValue(self, value):
        if not isinstance(value, (list, tuple)):
            value = [value]
        if not self.multi:
            assert len(value) == 1, "When multi is False, ToggleButtonArray value must be a single value"
        # Set corresponding button's value to be True and all others to be False
        for key, btn in self.buttons.items():
            btn.SetValue(key in value)
        # Restyle
        self._applyAppTheme()
        # Emit event
        evt = wx.CommandEvent(wx.EVT_CHOICE.typeId)
        evt.SetEventObject(self)
        wx.PostEvent(self, evt)

    def GetValue(self):
        # Return key of button(s) whose value is True
        values = []
        for key, btn in self.buttons.items():
            if btn.GetValue():
                values.append(key)
        # If single select, only return first value
        if not self.multi:
            values = values[0]
        return values

    def _applyAppTheme(self, target=None):
        # Set panel background
        self.SetBackgroundColour(ThemeMixin.appColors['panel_bg'])
        # Use OnHover event to set buttons to their default colors
        for btn in self.buttons.values():
            btn.OnHover()


def sanitize(inStr):
    """
    Process a string to remove any sensitive information, i.e. OAUTH keys
    """
    # Key-value pairs of patterns with what to replace them with
    patterns = {
        "https\:\/\/oauth2\:[\d\w]{64}@gitlab\.pavlovia\.org\/.*\.git": "[[OAUTH key hidden]]" # Remove any oauth keys
    }
    # Replace each pattern
    for pattern, repl in patterns.items():
        inStr = re.sub(pattern, repl, inStr)

    return inStr



# -*- coding: utf-8 -*-
"""Classes for the color picker."""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
import sys

import wx
import wx.stc as stc
from .ui import ColorPickerDialog
from psychopy.colors import Color, colorNames
# from psychopy.localization import _translate

SLIDER_RES = 255  # resolution of the slider for color channels, leave alone!


class ColorPickerState:
    """Class representing the state of the color picker. This is used to
    provide persistence between multiple invocations of the dialog within a
    single session.

    Parameters
    ----------
    color : Color
        Last color selected by the user.
    space : int
        Last RGB colorspace specified.
    rgbInputMode : int
        RGB input mode selected by the user.

    """
    def __init__(self, color, space, rgbInputMode):
        self.color = color
        self.space = space
        self.rgbInputMode = rgbInputMode


# Store the state of the color picker here for persistance. Here we define the
# defaults for when the colorpicker is first invoked.
COLORPICKER_STATE = ColorPickerState(
    color=Color((0, 0, 0, 1), space='rgba'),
    space=0,
    rgbInputMode=0  # PsychopPy RGB
)


class PsychoColorPicker(ColorPickerDialog):
    """Class for the color picker dialog.

    This dialog is used to standardize color selection across platforms. It also
    supports PsychoPy's RGB representation directly.

    This is a subclass of the auto-generated `ColorPickerDialog` class.

    Parameters
    ----------
    parent : object
        Reference to a :class:`~wx.Frame` which owns this dialog.

    """
    # NB: This is a *really* complicated dialog box as controls can update each
    # other and colors have many different formats that can be used to set
    # values. If you're not careful, one control updating the values of another
    # can trigger events which in-turn set the original control. This can lead
    # to deadlocks that may either stop the control from being updated or cause
    # an application error. There are probably better ways of doing things here
    # but beware of these potential pitfalls when making changes to this class.
    # -- mdc
    #
    def __init__(self, parent, context=None, allowInsert=True, allowCopy=True):
        ColorPickerDialog.__init__(self, parent)

        self.parent = parent
        self.allowInsert = allowInsert

        if not self.allowInsert:
            self.cmdInsert.Disable()

        self.allowCopy = allowCopy

        if not self.allowCopy:
            self.cmdCopy.Disable()
            self.cmdCopy.Hide()
            # make insert show 'OK' if the only option
            self.cmdInsert.Label = u'&OK'

        self.SetSizeHints(wx.DefaultSize, wx.DefaultSize)
        self.SetSize(wx.Size(640, 480))
        self.SetMinSize(wx.Size(640, 480))

        # store context
        self.context = context

        # output spaces mapped to the `cboOutputSpace` object
        self._spaces = {
            0: 'rgb',
            1: 'rgb1',
            2: 'rgb255',
            3: 'hex',
            4: 'hsv'
        }
        # what to show in the output selection, maps to the spaces above
        self._outputChoices = [
            u'PsychoPy RGB (rgb)',
            u'Normalized RGB (rgb1)',
            u'8-bit RGB (rgb255)',
            u'Hex/HTML (hex)',
            u'Hue-Saturation-Value (hsv)'
        ]

        # set preset color names, all except None
        self.lstColorPresets.SetItems(
            [col for col in colorNames.keys() if col is not None])

        # Functions to convert slider units to an RGB format to display in the
        # double-spin controls beside them.
        self._posToValFunc = {
            0: lambda v: 2 * (v / SLIDER_RES) - 1,  # [-1:1]
            1: lambda v: v / SLIDER_RES,  # [0:1]
            2: lambda v: v}  # [0:255]

        # inverse of the above functions, converts values to positions
        self._valToPosFunc = {
            0: lambda p: int(SLIDER_RES * (p + 1) / 2.),  # [-1:1]
            1: lambda p: int(p * SLIDER_RES),  # [0:1]
            2: lambda p: int(p)}  # [0:255]

        # initialize the window controls
        midpoint = int(SLIDER_RES / 2.)
        self.sldRedChannel.SetValue(midpoint)
        self.sldGreenChannel.SetValue(midpoint)
        self.sldBlueChannel.SetValue(midpoint)
        self.sldRedChannel.SetMax(SLIDER_RES)
        self.sldGreenChannel.SetMax(SLIDER_RES)
        self.sldBlueChannel.SetMax(SLIDER_RES)
        self.sldHueChannel.SetRange(0.0, 360.0)
        self.sldStaturationChannel.SetRange(0, SLIDER_RES)
        self.sldValueChannel.SetRange(0, SLIDER_RES)

        # current output color, should be displayed in the preview
        global COLORPICKER_STATE
        self._color = None
        self.color = COLORPICKER_STATE.color.copy()
        self.rgbInputMode = COLORPICKER_STATE.rgbInputMode

        # set the available output spaces
        self.cboOutputSpace.SetItems(self._outputChoices)
        self.cboOutputSpace.SetSelection(COLORPICKER_STATE.space)

    @property
    def color(self):
        """Current color the user has specified. Should be reflected in the
        preview area. Value has type :class:`psychopy.colors.Color`.

        This is the primary setter for the dialog's color value. It will
        automatically update all color space pages to reflect this color value.
        Pages should absolutely not attempt to set other page's values directly
        or risk blowing up the call stack.

        """
        return self._color

    @color.setter
    def color(self, value):
        if value is None:
            value = Color((0., 0., 0., 1.), space='rgba')

        self._color = value
        self.updateRGBPage()
        self.updateHSVPage()
        self.updateDialog()

    @property
    def rgbInputMode(self):
        """The RGB input mode (`int`). This will update how RGB channels are
        formatted when set.

        """
        # replaces the older radio box which returned indices
        if self.rdoRGBModePsychoPy.GetValue():
            return 0
        elif self.rdoRGBModeNormalized.GetValue():
            return 1
        elif self.rdoRGBMode255.GetValue():
            return 2
        else:
            return -1

    @rgbInputMode.setter
    def rgbInputMode(self, value):
        value = int(value)
        if value == 0:
            self.rdoRGBModePsychoPy.SetValue(1)
        elif value == 1:
            self.rdoRGBModeNormalized.SetValue(1)
        elif value == 2:
            self.rdoRGBMode255.SetValue(1)
        else:
            raise IndexError('Invalid index for `rgbInputMode`.')

        self.updateRGBChannels()
        self.updateDialog()

    # --------------------------------------------------------------------------
    # Dialog controls update methods
    #

    def updateDialog(self):
        """Update the color specified by the dialog. Call this whenever the
        color controls are updated by the user.

        """
        spinVals = [
            self.spnRedChannel.GetValue(),
            self.spnGreenChannel.GetValue(),
            self.spnBlueChannel.GetValue()]

        if self.rgbInputMode == 0:
            self.color.rgb = spinVals
        elif self.rgbInputMode == 1:
            self.color.rgb1 = spinVals
        elif self.rgbInputMode == 2:
            self.color.rgb255 = spinVals
        else:
            raise ValueError('Color channel format not supported.')

        # update the HSV page
        previewRGB = self.color.rgb255
        r, g, b = previewRGB
        self.pnlColorPreview.SetBackgroundColour(
            wx.Colour(r, g, b, alpha=wx.ALPHA_OPAQUE))
        self.pnlColorPreview.Refresh()

        self.txtHexRGB.SetValue(self.color.hex)

    def updateHSVPage(self):
        """Update values on the HSV page. This is called when some other page is
        used to pick the color and the HSV page needs to reflect that.

        """
        # get colors and convert to format wxPython controls can accept
        hsvColor = self.color.hsv
        self.sldHueChannel.SetValue(hsvColor[0])
        self.sldStaturationChannel.SetValue(hsvColor[1] * SLIDER_RES)
        self.sldValueChannel.SetValue(hsvColor[2] * SLIDER_RES)

        # set the value in the new range
        self.spnHueChannel.SetValue(hsvColor[0])
        self.spnSaturationChannel.SetValue(hsvColor[1])
        self.spnValueChannel.SetValue(hsvColor[2])

    def updateRGBPage(self):
        """Update values on the RGB page. This is called when some other page is
        used to pick the color and the RGB page needs to reflect that.

        """
        rgb255 = [int(i) for i in self.color.rgb255]
        self.sldRedChannel.SetValue(rgb255[0])
        self.sldGreenChannel.SetValue(rgb255[1])
        self.sldBlueChannel.SetValue(rgb255[2])
        # self.sldAlpha.SetValue(
        #    rgbaColor.alpha * SLIDER_RES)  # arrrg! should be 255!!!

        # set the value in the new range
        if self.rgbInputMode == 0:
            spnColVals = self.color.rgb
        elif self.rgbInputMode == 1:
            spnColVals = self.color.rgb1
        elif self.rgbInputMode == 2:
            spnColVals = self.color.rgb255
        else:
            raise ValueError(
                "Unknown RGB channel format specified. Did you add a new mode?")

        self.spnRedChannel.SetValue(spnColVals[0])
        self.spnGreenChannel.SetValue(spnColVals[1])
        self.spnBlueChannel.SetValue(spnColVals[2])

    def updateRGBChannels(self):
        """Update the values of the indicated RGB channels to reflect the
        selected color input mode.

        The RGB page has modes which changes the format of the input values.
        Calling this updates the controls to reflect the format.

        """
        # get colors and convert to format wxPython controls can accept
        channelMode = self.rgbInputMode
        convFunc = self._posToValFunc[channelMode]

        # update spinner values/ranges for each channel
        for spn in (self.spnRedChannel, self.spnGreenChannel, self.spnBlueChannel):
            spn.SetDigits(
                0 if channelMode == 2 else 4)
            spn.SetIncrement(
                1 if channelMode == 2 else 0.05)
            spn.SetMin(convFunc(0))
            spn.SetMax(convFunc(SLIDER_RES))

        self.updateRGBPage()

    # --------------------------------------------------------------------------
    # Events for RGB controls
    #

    def OnRedScroll(self, event):
        """Called when the red (RGB) channel slider is moved.
        """
        self.spnRedChannel.SetValue(
            self._posToValFunc[self.rgbInputMode](
                event.Position))
        self.updateHSVPage()
        self.updateDialog()

    def OnGreenScroll(self, event):
        """Called when the green (RGB) channel slider is moved.
        """
        self.spnGreenChannel.SetValue(
            self._posToValFunc[self.rgbInputMode](
                event.Position))
        self.updateHSVPage()
        self.updateDialog()

    def OnBlueScroll(self, event):
        """Called when the blue (RGB) channel slider is moved.
        """
        self.spnBlueChannel.SetValue(
            self._posToValFunc[self.rgbInputMode](
                event.Position))
        self.updateHSVPage()
        self.updateDialog()

    def OnRedSpin(self, event):
        """Called when the red (RGB) spin control is changed.
        """
        self.sldRedChannel.SetValue(
            self._valToPosFunc[self.rgbInputMode](event.Value))
        self.updateHSVPage()
        self.updateDialog()

    def OnGreenSpin(self, event):
        """Called when the green (RGB) spin control is changed.
        """
        self.sldGreenChannel.SetValue(
            self._valToPosFunc[self.rgbInputMode](event.Value))
        self.updateHSVPage()
        self.updateDialog()

    def OnBlueSpin(self, event):
        """Called when the blue (RGB) spin control is changed.
         """
        self.sldBlueChannel.SetValue(
            self._valToPosFunc[self.rgbInputMode](event.Value))
        self.updateHSVPage()
        self.updateDialog()

    def OnRGBModePsychoPy(self, event):
        """Called when the user selects 'PsychoPy RGB' mode as the RGB input
        format.
        """
        self.updateRGBChannels()

    def OnRGBModeNormalized(self, event):
        """Called when the user selects 'Normalized RGB' mode as the RGB input
        format.
        """
        self.updateRGBChannels()

    def OnRGBMode255(self, event):
        """Called when the user selects '8-Bit RGB' mode as the RGB input
        format.
        """
        self.updateRGBChannels()

    # --------------------------------------------------------------------------
    # Events for HSV controls
    #

    def OnHueSpin(self, event):
        """Called when the hue (HSV) spin control is changed.
        """
        self.sldHueChannel.SetValue(event.GetValue())
        self.color.hsv = (
            self.spnHueChannel.GetValue(),
            self.spnSaturationChannel.GetValue(),
            self.spnValueChannel.GetValue())
        self.updateRGBPage()
        self.updateDialog()

    def OnHueScroll(self, event):
        """Called when the hue (HSV) channel slider is moved.
        """
        self.spnHueChannel.SetValue(event.Position)
        self.color.hsv = (
            self.spnHueChannel.GetValue(),
            self.spnSaturationChannel.GetValue(),
            self.spnValueChannel.GetValue())
        self.updateRGBPage()
        self.updateDialog()

    def OnSaturationSpin(self, event):
        """Called when the saturation (HSV) spin control is changed.
        """
        self.sldStaturationChannel.SetValue(event.GetValue() * SLIDER_RES)
        self.color.hsv = (
            self.spnHueChannel.GetValue(),
            self.spnSaturationChannel.GetValue(),
            self.spnValueChannel.GetValue())
        self.updateRGBPage()
        self.updateDialog()

    def OnSaturationScroll(self, event):
        """Called when the saturation (HSV) channel slider is moved.
        """
        self.spnSaturationChannel.SetValue(event.Position / SLIDER_RES)
        self.color.hsv = (
            self.spnHueChannel.GetValue(),
            self.spnSaturationChannel.GetValue(),
            self.spnValueChannel.GetValue())
        self.updateRGBPage()
        self.updateDialog()

    def OnValueSpin(self, event):
        """Called when the value (HSV) spin control is changed.
        """
        self.sldValueChannel.SetValue(event.GetValue() * SLIDER_RES)
        self.color.hsv = (
            self.spnHueChannel.GetValue(),
            self.spnSaturationChannel.GetValue(),
            self.spnValueChannel.GetValue())
        self.updateRGBPage()
        self.updateDialog()

    def OnValueScroll(self, event):
        """Called when the value (HSV) channel slider is moved.
        """
        self.spnValueChannel.SetValue(event.Position / SLIDER_RES)
        self.color.hsv = (
            self.spnHueChannel.GetValue(),
            self.spnSaturationChannel.GetValue(),
            self.spnValueChannel.GetValue())
        self.updateRGBPage()
        self.updateDialog()

    # --------------------------------------------------------------------------
    # Hex RGB input methods
    #
    def OnHexRGBKeyDown(self, event):
        """Called when the user manually enters a hex value into the field.

        If the color value is valid, the new value will appear and the channels
        will update. If not, the box will revert back to the last valid value.

        """
        # keydown need to be processed like this on MacOS
        if event.GetKeyCode() == wx.WXK_RETURN:
            oldHexColor = self.color.hex
            try:
                self.color.rgba = Color(
                    self.txtHexRGB.GetValue(), space='hex').rgba
                self.updateRGBPage()
                self.updateHSVPage()
                self.updateDialog()
            except ValueError:
                self.txtHexRGB.SetValue(oldHexColor)
        else:
            event.Skip()

    def updateHex(self):
        """Update the hex/HTML value using the color specified by the dialog.
        """
        self.txtHexRGB.SetValue(self.color.hex)

    # --------------------------------------------------------------------------
    # Color preset list methods
    #
    def OnPresetSelect(self, event):
        """Called when the user selects a preset from the list.
        """
        idxSelection = self.lstColorPresets.GetSelection()
        if idxSelection == -1:  # event occurred with no valid selection
            event.Skip()

        presetColorKey = self.lstColorPresets.GetString(idxSelection)
        presetColor = Color(colorNames[presetColorKey], space='rgb')

        self.color = presetColor

    # --------------------------------------------------------------------------
    # Output methods
    #
    def getOutputValue(self):
        """Get the string value using the specified output format.

        Returns
        -------
        str
            Color value using the current output format.

        """
        outputSpace = self.cboOutputSpace.GetSelection()
        dlgCol = self.GetTopLevelParent().color
        if outputSpace == 0:  # RGB
            colorOut = '{:.4f}, {:.4f}, {:.4f}'.format(*dlgCol.rgb)
        elif outputSpace == 1:  # RGB1
            colorOut = '{:.4f}, {:.4f}, {:.4f}'.format(*dlgCol.rgb1)
        elif outputSpace == 2:  # RGB255
            colorOut = '{:d}, {:d}, {:d}'.format(
                *[int(i) for i in dlgCol.rgb255])
        elif outputSpace == 3:  # Hex
            colorOut = "'{}'".format(dlgCol.hex)
        elif outputSpace == 4:  # HSV
            colorOut = '{:.4f}, {:.4f}, {:.4f}'.format(*dlgCol.hsv)
        else:
            raise ValueError(
                "Invalid output color space selection. Have you added any "
                "choices to `cboOutputSpace`?")

        return colorOut

    def _copyToClipboard(self, text):
        """Copy text to the clipboard.

        Shows an error dialog if the clipboard cannot be opened to inform the
        user the values have not been copied.

        Parameters
        ----------
        text : str
            Text to copy to the clipboard.

        """
        # copy the value to the clipboard
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(wx.TextDataObject(str(text)))
            wx.TheClipboard.Close()
        else:
            # Raised if the clipboard fails to open, warns the user the value
            # wasn't copied.
            errDlg = wx.MessageDialog(
                self,
                'Failed to open the clipboard, output value not copied.',
                'Clipboard Error',
                wx.OK | wx.ICON_ERROR)
            errDlg.ShowModal()
            errDlg.Destroy()

    def _saveState(self):
        """Call this to make dialog value persistent across invocations.
        """
        global COLORPICKER_STATE
        COLORPICKER_STATE.color = self.color.copy()
        COLORPICKER_STATE.space = self.cboOutputSpace.GetSelection()
        COLORPICKER_STATE.rgbInputMode = self.rgbInputMode

    def OnInsert(self, event):
        """Event to copy the color to the clipboard as a an object.
        """
        if not self.allowInsert:
            event.Skip()

        if isinstance(self.context, wx.TextCtrl):
            self.context.SetValue(self.getOutputValue())
        elif isinstance(self.context, stc.StyledTextCtrl):
            self.context.InsertText(
                self.context.GetCurrentPos(), "(" + self.getOutputValue() + ")")

        self._saveState()  # retain state
        self.Close()

    def OnCopy(self, event):
        """Event to copy the color to the clipboard as a value.
        """
        if not self.allowCopy:
            event.Skip()

        self._copyToClipboard(self.getOutputValue())  # copy out to clipboard

        self._saveState()
        self.Close()

    def OnCancel(self, event):
        """Called when the cancel button is clicked.
        """
        self.Close()

    # def OnClose(self, event):
    #     """Called when the 'x' is clicked on the title bar."""
    #     self._saveState()
    #     self.Close()


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright (C) 2014 Allen Institute for Brain Science

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License Version 3
as published by the Free Software Foundation on 29 June 2007.
This program is distributed WITHOUT WARRANTY OF MERCHANTABILITY OR FITNESS
FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESSED OR IMPLIED.
See the GNU General Public License Version 3 for more details.
You should have received a copy of the GNU General Public License along
with this program. If not, see http://www.gnu.org/licenses/
"""

import ctypes
import numpy as np
from psychopy import logging
import pyglet
GL = pyglet.gl


class Warper:
    """Class to perform warps.

    Supports spherical, cylindrical, warpfile, or None (disabled) warps
    """

    def __init__(self,
                 win,
                 warp=None,
                 warpfile=None,
                 warpGridsize=300,
                 eyepoint=(0.5, 0.5),
                 flipHorizontal=False,
                 flipVertical=False):
        """Warping is a final operation which can be optionally performed on
        each frame just before transmission to the display. It is useful
        for perspective correction when the eye to monitor distance is
        small (say, under 50 cm), or when projecting to domes or other
        non-planar surfaces.

        These attributes define the projection and can be altered
        dynamically using the changeProjection() method.

        :Parameters:
            win : Handle to the window.

            warp : 'spherical', 'cylindrical, 'warpfile' or *None*
                This table gives the main properties of each projection

                +-----------+---------------+-----------+------------+--------------------+
                | Warp      | eyepoint      | verticals | horizontals| perspective correct|
                |           | modifies warp | parallel  | parallel   |                    |
                +===========+===============+===========+============+====================+
                |spherical  |    y          |   n       |  n         |   y                |
                +-----------+---------------+-----------+------------+--------------------+
                |cylindrical|    y          |   y       |  n         |   n                |
                +-----------+---------------+-----------+------------+--------------------+
                | warpfile  |    n          |   ?       |  ?         |   ?                |
                +-----------+---------------+-----------+------------+--------------------+
                | None      |    n          |   y       |  y         |   n                |
                +-----------+---------------+-----------+------------+--------------------+

            warpfile : *None* or filename containing Blender and Paul Bourke
                compatible warp definition.
                (see http://paulbourke.net/dome/warpingfisheye/)
            warpGridsize : 300
                Defines the resolution of the warp in both X and Y when
                not using a warpfile. Typical values would be 64-300
                trading off tolerance for jaggies for speed.
            eyepoint : [0.5, 0.5] center of the screen
                Position of the eye in X and Y as a fraction of the
                normalized screen width and height.
                [0,0] is the bottom left of the screen.
                [1,1] is the top right of the screen.
            flipHorizontal: True or *False*
                Flip the entire output horizontally.  Useful for back
                projection scenarious.
            flipVertical: True or *False*
                Flip the entire output vertically. useful if projector is
                flipped upside down.

        :notes:
            1) The eye distance from the screen is initialized from the
                monitor definition.
            2) The eye distance can be altered dynamically by changing
                'warper.dist_cm' and then calling changeProjection().

        Example usage to create a spherical projection::

            from psychopy.visual.windowwarp import Warper
            win = Window(monitor='testMonitor', screen=1,
                         fullscr=True, useFBO = True)
            warper = Warper(win,
                            warp='spherical',
                            warpfile = "",
                            warpGridsize = 128,
                            eyepoint = [0.5, 0.5],
                            flipHorizontal = False,
                            flipVertical = False)

        """
        super(Warper, self).__init__()
        self.win = win
        # monkey patch the warp method
        win._renderFBO = self.drawWarp
        self.warp = warp
        self.warpfile = warpfile
        self.warpGridsize = warpGridsize
        self.eyepoint = eyepoint
        self.flipHorizontal = flipHorizontal
        self.flipVertical = flipVertical
        self.initDefaultWarpSize()

        #   get the eye distance from the monitor object,
        #   but the pixel dimensions from the actual window object
        w, h = win.size
        self.aspect = w/h
        self.dist_cm = win.monitor.getDistance()
        if self.dist_cm is None:
            # create a fake monitor if one isn't defined
            self.dist_cm = 30.0
            self.mon_width_cm = 50.0
            logging.warning('Monitor is not calibrated')
        else:
            self.mon_width_cm = win.monitor.getWidth()
        self.mon_height_cm = self.mon_width_cm / self.aspect
        self.mon_width_pix = w
        self.mon_height_pix = h
        self.changeProjection(self.warp, self.warpfile, self.eyepoint)

    def drawWarp(self):
        """Warp the output, using the vertex, texture, and optionally an
        opacity array.
        """
        GL.glUseProgram(0)
        GL.glColorMask(True, True, True, True)

        # point to color (opacity)
        if self.gl_color is not None:
            GL.glEnableClientState(GL.GL_COLOR_ARRAY)
            GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.gl_color)
            GL.glColorPointer(4, GL.GL_FLOAT, 0, None)
            GL.glEnable(GL.GL_BLEND)
            GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ZERO)

        # point to vertex data
        GL.glEnableClientState(GL.GL_VERTEX_ARRAY)
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.gl_vb)
        GL.glVertexPointer(2, GL.GL_FLOAT, 0, None)

        # point to texture
        GL.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY)
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.gl_tb)
        GL.glTexCoordPointer(2, GL.GL_FLOAT, 0, None)

        # draw quads
        GL.glDrawArrays(GL.GL_QUADS, 0, self.nverts)

        # cleanup
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, 0)
        GL.glDisableClientState(GL.GL_VERTEX_ARRAY)
        GL.glDisableClientState(GL.GL_TEXTURE_COORD_ARRAY)

        if self.gl_color is not None:
            GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
            GL.glDisableClientState(GL.GL_COLOR_ARRAY)

    def initDefaultWarpSize(self):
        self.xgrid = self.warpGridsize
        self.ygrid = self.warpGridsize

    def changeProjection(self, warp, warpfile=None, eyepoint=(0.5, 0.5),
                         flipHorizontal=False, flipVertical=False):
        """Allows changing the warp method on the fly.  Uses the same
        parameter definitions as constructor.
        """
        self.warp = warp
        self.warpfile = warpfile
        self.eyepoint = list(eyepoint)
        self.flipHorizontal = flipHorizontal
        self.flipVertical = flipVertical

        # warpfile might have changed the size...
        self.initDefaultWarpSize()

        if self.warp is None:
            self.projectionNone()
        elif self.warp == 'spherical':
            self.projectionSphericalOrCylindrical(False)
        elif self.warp == 'cylindrical':
            self.projectionSphericalOrCylindrical(True)
        elif self.warp == 'warpfile':
            self.projectionWarpfile()
        else:
            raise ValueError('Unknown warp specification: %s' % self.warp)

    def projectionNone(self):
        """No warp, same projection as original PsychoPy
        """
        # Vertex data
        v0 = (-1.0, -1.0)
        v1 = (-1.0, 1.0)
        v2 = (1.0, 1.0)
        v3 = (1.0, -1.0)

        # Texture coordinates
        t0 = (0.0, 0.0)
        t1 = (0.0, 1.0)
        t2 = (1.0, 1.0)
        t3 = (1.0, 0.0)

        vertices = np.array([v0, v1, v2, v3], 'float32')
        tcoords = np.array([t0, t1, t2, t3], 'float32')

        # draw four quads during rendering loop
        self.nverts = 4
        self.createVertexAndTextureBuffers(vertices, tcoords)

    def projectionSphericalOrCylindrical(self, isCylindrical=False):
        """Correct perspective on flat screen using either a spherical or
        cylindrical projection.
        """
        self.nverts = (self.xgrid - 1) * (self.ygrid - 1) * 4

        # eye position in cm
        xEye = self.eyepoint[0] * self.mon_width_cm
        yEye = self.eyepoint[1] * self.mon_height_cm

        # create vertex grid array, and texture coords
        # times 4 for quads
        vertices = np.zeros(
            ((self.xgrid - 1) * (self.ygrid - 1) * 4, 2), dtype='float32')
        tcoords = np.zeros(
            ((self.xgrid - 1) * (self.ygrid - 1) * 4, 2), dtype='float32')

        equalDistanceX = np.linspace(0, self.mon_width_cm, self.xgrid)
        equalDistanceY = np.linspace(0, self.mon_height_cm, self.ygrid)

        # vertex coordinates
        x_c = np.linspace(-1.0, 1.0, self.xgrid)
        y_c = np.linspace(-1.0, 1.0, self.ygrid)
        x_coords, y_coords = np.meshgrid(x_c, y_c)

        x = np.zeros(((self.xgrid), (self.ygrid)), dtype='float32')
        y = np.zeros(((self.xgrid), (self.ygrid)), dtype='float32')

        x[:, :] = equalDistanceX - xEye
        y[:, :] = equalDistanceY - yEye
        y = np.transpose(y)

        r = np.sqrt(np.square(x) + np.square(y) + np.square(self.dist_cm))

        azimuth = np.arctan(x / self.dist_cm)
        altitude = np.arcsin(y / r)

        # calculate the texture coordinates
        if isCylindrical:
            tx = self.dist_cm * np.sin(azimuth)
            ty = self.dist_cm * np.sin(altitude)
        else:
            tx = self.dist_cm * (1 + x/r) - self.dist_cm
            ty = self.dist_cm * (1 + y/r) - self.dist_cm

        # prevent div0
        azimuth[azimuth == 0] = np.finfo(np.float32).eps
        altitude[altitude == 0] = np.finfo(np.float32).eps

        # the texture coordinates (which are now lying on the sphere)
        # need to be remapped back onto the plane of the display.
        # This effectively stretches the coordinates away from the eyepoint.

        if isCylindrical:
            tx = tx * azimuth / np.sin(azimuth)
            ty = ty * altitude / np.sin(altitude)
        else:
            centralAngle = np.arccos(
                np.cos(altitude) * np.cos(np.abs(azimuth)))
            # distance from eyepoint to texture vertex
            arcLength = centralAngle * self.dist_cm
            # remap the texture coordinate
            theta = np.arctan2(ty, tx)
            tx = arcLength * np.cos(theta)
            ty = arcLength * np.sin(theta)

        u_coords = tx / self.mon_width_cm + 0.5
        v_coords = ty / self.mon_height_cm + 0.5

        # loop to create quads
        vdex = 0
        for y in range(0, self.ygrid - 1):
            for x in range(0, self.xgrid - 1):
                index = y * (self.xgrid) + x

                vertices[vdex + 0, 0] = x_coords[y, x]
                vertices[vdex + 0, 1] = y_coords[y, x]
                vertices[vdex + 1, 0] = x_coords[y, x + 1]
                vertices[vdex + 1, 1] = y_coords[y, x + 1]
                vertices[vdex + 2, 0] = x_coords[y + 1, x + 1]
                vertices[vdex + 2, 1] = y_coords[y + 1, x + 1]
                vertices[vdex + 3, 0] = x_coords[y + 1, x]
                vertices[vdex + 3, 1] = y_coords[y + 1, x]

                tcoords[vdex + 0, 0] = u_coords[y, x]
                tcoords[vdex + 0, 1] = v_coords[y, x]
                tcoords[vdex + 1, 0] = u_coords[y, x + 1]
                tcoords[vdex + 1, 1] = v_coords[y, x + 1]
                tcoords[vdex + 2, 0] = u_coords[y + 1, x + 1]
                tcoords[vdex + 2, 1] = v_coords[y + 1, x + 1]
                tcoords[vdex + 3, 0] = u_coords[y + 1, x]
                tcoords[vdex + 3, 1] = v_coords[y + 1, x]

                vdex += 4
        self.createVertexAndTextureBuffers(vertices, tcoords)

    def projectionWarpfile(self):
        """Use a warp definition file to create the projection.
            See: http://paulbourke.net/dome/warpingfisheye/
        """
        try:
            fh = open(self.warpfile)
            lines = fh.readlines()
            fh.close()
            filetype = int(lines[0])
            rc = list(map(int, lines[1].split()))
            cols, rows = rc[0], rc[1]
            warpdata = np.loadtxt(self.warpfile, skiprows=2)
        except Exception:
            error = 'Unable to read warpfile: ' + self.warpfile
            logging.warning(error)
            print(error)
            return

        if (cols * rows != warpdata.shape[0] or
                warpdata.shape[1] != 5 or
                filetype != 2):
            error = 'warpfile data incorrect: ' + self.warpfile
            logging.warning(error)
            print(error)
            return

        self.xgrid = cols
        self.ygrid = rows

        self.nverts = (self.xgrid - 1) * (self.ygrid - 1) * 4

        # create vertex grid array, and texture coords times 4 for quads
        vertices = np.zeros(
            ((self.xgrid - 1) * (self.ygrid - 1) * 4, 2), dtype='float32')
        tcoords = np.zeros(
            ((self.xgrid - 1) * (self.ygrid - 1) * 4, 2), dtype='float32')
        # opacity is RGBA
        opacity = np.ones(
            ((self.xgrid - 1) * (self.ygrid - 1) * 4, 4), dtype='float32')

        # loop to create quads
        vdex = 0
        for y in range(0, self.ygrid - 1):
            for x in range(0, self.xgrid - 1):
                index = y * (self.xgrid) + x

                vertices[vdex + 0, 0] = warpdata[index, 0]  # x_coords[y,x]
                vertices[vdex + 0, 1] = warpdata[index, 1]  # y_coords[y,x]
                # x_coords[y,x+1]
                vertices[vdex + 1, 0] = warpdata[index + 1, 0]
                # y_coords[y,x+1]
                vertices[vdex + 1, 1] = warpdata[index + 1, 1]
                # x_coords[y+1,x+1]
                vertices[vdex + 2, 0] = warpdata[index + cols + 1, 0]
                # y_coords[y+1,x+1]
                vertices[vdex + 2, 1] = warpdata[index + cols + 1, 1]
                # x_coords[y+1,x]
                vertices[vdex + 3, 0] = warpdata[index + cols, 0]
                # y_coords[y+1,x]
                vertices[vdex + 3, 1] = warpdata[index + cols, 1]

                # u_coords[y,x]
                tcoords[vdex + 0, 0] = warpdata[index, 2]
                # v_coords[y,x]
                tcoords[vdex + 0, 1] = warpdata[index, 3]
                # u_coords[y,x+1]:
                tcoords[vdex + 1, 0] = warpdata[index + 1, 2]
                # v_coords[y,x+1]:
                tcoords[vdex + 1, 1] = warpdata[index + 1, 3]
                # u_coords[y+1,x+1]:
                tcoords[vdex + 2, 0] = warpdata[index + cols + 1, 2]
                # v_coords[y+1,x+1]:
                tcoords[vdex + 2, 1] = warpdata[index + cols + 1, 3]
                # u_coords[y+1,x]
                tcoords[vdex + 3, 0] = warpdata[index + cols, 2]
                # v_coords[y+1,x]:
                tcoords[vdex + 3, 1] = warpdata[index + cols, 3]

                opacity[vdex, 3] = warpdata[index, 4]
                opacity[vdex + 1, 3] = warpdata[index + 1, 4]
                opacity[vdex + 2, 3] = warpdata[index + cols + 1, 4]
                opacity[vdex + 3, 3] = warpdata[index + cols, 4]

                vdex += 4
        self.createVertexAndTextureBuffers(vertices, tcoords, opacity)

    def createVertexAndTextureBuffers(self, vertices, tcoords, opacity=None):
        """Allocate hardware buffers for vertices, texture coordinates,
        and optionally opacity.
        """
        if self.flipHorizontal:
            vertices[:, 0] = -vertices[:, 0]
        if self.flipVertical:
            vertices[:, 1] = -vertices[:, 1]

        GL.glEnableClientState(GL.GL_VERTEX_ARRAY)

        # type and size for arrays
        arrType = ctypes.c_float
        ptrType = ctypes.POINTER(arrType)
        nbytes = ctypes.sizeof(arrType)

        # vertex buffer in hardware
        self.gl_vb = GL.GLuint()
        GL.glGenBuffers(1, self.gl_vb)
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.gl_vb)
        GL.glBufferData(
            GL.GL_ARRAY_BUFFER,
            vertices.size * nbytes,
            vertices.ctypes.data_as(ptrType),
            GL.GL_STATIC_DRAW)

        # vertex buffer texture data in hardware
        self.gl_tb = GL.GLuint()
        GL.glGenBuffers(1, self.gl_tb)
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.gl_tb)
        GL.glBufferData(
            GL.GL_ARRAY_BUFFER,
            tcoords.size * nbytes,
            tcoords.ctypes.data_as(ptrType),
            GL.GL_STATIC_DRAW)

        # opacity buffer in hardware (only for warp files)
        if opacity is not None:
            self.gl_color = GL.GLuint()
            GL.glGenBuffers(1, self.gl_color)
            GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.gl_color)

            # convert opacity to RGBA, one point for each corner of the quad
            GL.glBufferData(
                GL.GL_ARRAY_BUFFER,
                opacity.size * nbytes,
                opacity.ctypes.data_as(ptrType),
                GL.GL_STATIC_DRAW)

        else:
            self.gl_color = None

        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, 0)
        GL.glDisableClientState(GL.GL_VERTEX_ARRAY)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from pathlib import Path
from psychopy.experiment.components import Param, getInitVals, _translate, BaseVisualComponent
from psychopy.visual import form
from psychopy.localization import _localized as __localized
_localized = __localized.copy()

__author__ = 'Jon Peirce, David Bridges, Anthony Haffey'

# only use _localized values for label values, nothing functional:
_localized.update({'Items': _translate('Items'),
                   'Text Height': _translate('Text Height'),
                   'Style': _translate('Styles'),
                   'Item Padding': _translate('Item Padding'),
                   'Data Format': _translate('Data Format'),
                   'Randomize': _translate('Randomize')
                   })
knownStyles = list(form.Form.knownStyles)


class FormComponent(BaseVisualComponent):
    """A class for presenting a survey as a Builder component"""

    categories = ['Responses']
    targets = ['PsychoPy', 'PsychoJS']
    iconFile = Path(__file__).parent / 'form.png'
    tooltip = _translate('Form: a Psychopy survey tool')
    beta = True

    def __init__(self, exp, parentName,
                 name='form',
                 items='.csv',
                 textHeight=0.03,
                 font="Open Sans",
                 randomize=False,
                 fillColor='',
                 borderColor='',
                 itemColor='white',
                 responseColor='white',
                 markerColor='red',
                 size=(1, .7),
                 pos=(0, 0),
                 style='dark',
                 itemPadding=0.05,
                 startType='time (s)', startVal='0.0',
                 stopType='duration (s)', stopVal='',
                 startEstim='', durationEstim='',
                 # legacy
                 color='white'):

        super(FormComponent, self).__init__(
            exp, parentName, name=name,
            pos=pos, size=size,
            color=color, fillColor=fillColor, borderColor=borderColor,
            startType=startType, startVal=startVal,
            stopType=stopType, stopVal=stopVal,
            startEstim=startEstim, durationEstim=durationEstim)

        # these are defined by the BaseVisual but we don't want them
        del self.params['ori']
        del self.params['units']  # we only support height units right now
        del self.params['color']

        self.type = 'Form'
        self.url = "https://www.psychopy.org/builder/components/form.html"
        self.exp.requirePsychopyLibs(['visual', 'event', 'logging'])

        # params
        self.order += ['Items', 'Randomize',  # Basic tab
                       'Data Format',  # Data tab
                      ]
        self.order.insert(self.order.index("units"), "Item Padding")

        # normal params:
        # = the usual as inherited from BaseComponent plus:

        self.params['Items'] = Param(
            items, valType='file', inputType="table", allowedTypes=[], categ='Basic',
            updates='constant',
            hint=_translate("The csv filename containing the items for your survey."),
            label=_localized['Items'])

        self.params['Text Height'] = Param(
            textHeight, valType='num', inputType="single", allowedTypes=[], categ='Formatting',
            updates='constant',
            hint=_translate("The size of the item text for Form"),
            label=_localized['Text Height'])

        self.params['Font'] = Param(
            font, valType='str', inputType="single", allowedTypes=[], categ='Formatting',
            updates='constant', allowedUpdates=["constant"],
            hint=_translate("The font name (e.g. Comic Sans)"),
            label=_translate('Font'))

        self.params['Randomize'] = Param(
            randomize, valType='bool', inputType="bool", allowedTypes=[], categ='Basic',
            updates='constant',
            hint=_translate("Do you want to randomize the order of your questions?"),
            label=_localized['Randomize'])

        self.params['Item Padding'] = Param(
            itemPadding, valType='num', inputType="single", allowedTypes=[], categ='Layout',
            updates='constant',
            hint=_translate("The padding or space between items."),
            label=_localized['Item Padding'])

        self.params['Data Format'] = Param(
            'rows', valType='str', inputType="choice", allowedTypes=[], categ='Basic',
            allowedVals=['columns', 'rows'],
            updates='constant',
            hint=_translate("Store item data by columns, or rows"),
            label=_localized['Data Format'])

        # Appearance
        for param in ['fillColor', 'borderColor', 'itemColor', 'responseColor', 'markerColor', 'Style']:
            if param in self.order:
                self.order.remove(param)
            self.order.insert(
                self.order.index("colorSpace"),
                param
            )

        self.params['Style'] = Param(
            style, valType='str', inputType="choice", categ="Appearance",
            updates='constant', allowedVals=knownStyles + ["custom..."],
            hint=_translate(
                    "Styles determine the appearance of the form"),
            label=_localized['Style'])

        for param in ['fillColor', 'borderColor', 'itemColor', 'responseColor', 'markerColor']:
            self.depends += [{
                "dependsOn": "Style",  # must be param name
                "condition": "=='custom...'",  # val to check for
                "param": param,  # param property to alter
                "true": "enable",  # what to do with param if condition is True
                "false": "disable",  # permitted: hide, show, enable, disable
            }]

        self.params['fillColor'].hint = _translate("Color of the form's background")

        self.params['borderColor'].hint = _translate("Color of the outline around the form")

        self.params['itemColor'] = Param(itemColor,
            valType='color', inputType="color", categ='Appearance',
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=_translate("Base text color for questions"),
            label=_translate("Item Color"))

        self.params['responseColor'] = Param(responseColor,
            valType='color', inputType="color", categ='Appearance',
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=_translate("Base text color for responses, also sets color of lines in sliders and borders of textboxes"),
            label=_translate("Response Color"))

        self.params['markerColor'] = Param(markerColor,
            valType='color', inputType="color", categ='Appearance',
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=_translate("Color of markers and the scrollbar"),
            label=_translate("Marker Color"))

        self.params['pos'].allowedUpdates = []
        self.params['size'].allowedUpdates = []

    def writeInitCode(self, buff):
        inits = getInitVals(self.params)
        # build up an initialization string for Form():
        code = (
            "win.allowStencil = True\n"
            "%(name)s = visual.Form(win=win, name='%(name)s',\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
            "items=%(Items)s,\n"
            "textHeight=%(Text Height)s,\n"
            "font=%(Font)s,\n"
            "randomize=%(Randomize)s,\n"
            "style=%(Style)s,\n"
            "fillColor=%(fillColor)s, borderColor=%(borderColor)s, itemColor=%(itemColor)s, \n"
            "responseColor=%(responseColor)s, markerColor=%(markerColor)s, colorSpace=%(colorSpace)s, \n"
            "size=%(size)s,\n"
            "pos=%(pos)s,\n"
            "itemPadding=%(Item Padding)s"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
            ")\n"
        )
        buff.writeIndentedLines(code % inits)

    def writeInitCodeJS(self, buff):
        inits = getInitVals(self.params)
        # build up an initialization string for Form():
        initStr = ("{name} = new visual.Form({{\n"
                   "  win : psychoJS.window, name:'{name}',\n"
                   "  items : {Items},\n"
                   "  textHeight : {Text Height},\n"
                   "  font : {Font},\n"
                   "  randomize : {Randomize},\n"
                   "  size : {size},\n"
                   "  pos : {pos},\n"
                   "  style : {Style},\n"
                   "  itemPadding : {Item Padding}\n"
                   "}});\n".format(**inits))
        buff.writeIndentedLines(initStr)

    def writeRoutineEndCode(self, buff):
        # save data, according to row/col format
        buff.writeIndented("{name}.addDataToExp(thisExp, {Data Format})\n"
                           .format(**self.params))
        buff.writeIndented("{name}.autodraw = False\n"
                           .format(**self.params))

    def writeRoutineEndCodeJS(self, buff):
        # save data, according to row/col format
        buff.writeIndented("{name}.addDataToExp(psychoJS.experiment, {Data Format});\n"
                           .format(**self.params))



import re
from ..utils import _q, _lb, _rb, _d, _sl
from psychopy.experiment import Param, utils as exputils


def test_param_str():
    """
    Test that params convert to str as expected in both Python and JS
    """

    cases = [
        # Regular string
        {"obj": Param("Hello there", "str"),
         "py": f"{_q}Hello there{_q}",
         "js": f"{_q}Hello there{_q}"},
        # Enforced string
        {"obj": Param("\\, | or /", "str", canBePath=False),
         "py": f"{_q}{_sl}, | or /{_q}",
         "js": f"{_q}{_sl}, | or /{_q}"},
        # Dollar string
        {"obj": Param("$win.color", "str"),
         "py": f"win.color",
         "js": f"psychoJS.window.color"},
        # Integer
        {"obj": Param("1", "int"),
         "py": f"1",
         "js": f"1"},
        # Float
        {"obj": Param("1", "num"),
         "py": f"1.0",
         "js": f"1.0"},
        # File path
        {"obj": Param("C://Downloads//file.ext", "file"),
         "py": f"{_q}C:/Downloads/file.ext{_q}",
         "js": f"{_q}C:/Downloads/file.ext{_q}"},
        # Table path
        {"obj": Param("C://Downloads//file.csv", "table"),
         "py": f"{_q}C:/Downloads/file.csv{_q}",
         "js": f"{_q}C:/Downloads/file.csv{_q}"},
        # Color
        {"obj": Param("red", "color"),
         "py": f"{_q}red{_q}",
         "js": f"{_q}red{_q}"},
        # RGB Color
        {"obj": Param("0.7, 0.7, 0.7", "color"),
         "py": f"{_lb}0.7, 0.7, 0.7{_rb}",
         "js": f"{_lb}0.7, 0.7, 0.7{_rb}"},
        # Code
        {"obj": Param("win.color", "code"),
         "py": f"win.color",
         "js": f"psychoJS.window.color"},
        # Extended code
        {"obj": Param("for x in y:\n\tprint(y)", "extendedCode"),
         "py": f"for x in y:\n\tprint(y)",
         "js": f"for x in y:\n\tprint(y)"},  # this will change when snipped2js is fully working
        # List
        {"obj": Param("1, 2, 3", "list"),
         "py": f"{_lb}1, 2, 3{_rb}",
         "js": f"{_lb}1, 2, 3{_rb}"},
        # Extant file path marked as str
        {"obj": Param(__file__, "str"),
         "py": f"{_q}{__file__.replace(_sl, '/')}{_q}",
         "js": f"{_q}{__file__.replace(_sl, '/')}{_q}"},
        # Nonexistent file path marked as str
        {"obj": Param("C:\\\\Downloads\\file.csv", "str"),
         "py": f"{_q}C:/Downloads/file.csv{_q}",
         "js": f"{_q}C:/Downloads/file.csv{_q}"},
        # Underscored file path marked as str
        {"obj": Param("C:\\\\Downloads\\_file.csv", "str"),
         "py": f"{_q}C:/Downloads/_file.csv{_q}",
         "js": f"{_q}C:/Downloads/_file.csv{_q}"},
        # Escaped $ in str
        {"obj": Param("This costs \\$4.20", "str"),
         "py": f"{_q}This costs $4.20{_q}",
         "js": f"{_q}This costs $4.20{_q}"},
        # Unescaped \ in str
        {"obj": Param("This \\ that", "str"),
         "py": f"{_q}This \\\\ that{_q}",
         "js": f"{_q}This \\\\ that{_q}"},
        # Name containing "var" (should no longer return blank as of #4336)
        {"obj": Param("variableName", "code"),
         "py": f"variableName",
         "js": f"variableName"},
        # Color param with a $
        {"obj": Param("$letterColor", "color"),
         "py": f"letterColor",
         "js": f"letterColor"},
        # Double quotes naked list
        {'obj': Param("\"left\", \"down\", \"right\"", "list"),
         'py': f"{_lb}{_q}left{_q}, {_q}down{_q}, {_q}right{_q}{_rb}",
         'js': f"{_lb}{_q}left{_q}, {_q}down{_q}, {_q}right{_q}{_rb}"},
        # Single quotes naked list
        {'obj': Param("\'left\', \'down\', \'right\'", "list"),
         'py': f"{_lb}{_q}left{_q}, {_q}down{_q}, {_q}right{_q}{_rb}",
         'js': f"{_lb}{_q}left{_q}, {_q}down{_q}, {_q}right{_q}{_rb}"},
        # Single quotes tuple syntax
        {'obj': Param("(\'left\', \'down\', \'right\')", "list"),
         'py': f"{_lb}{_q}left{_q}, {_q}down{_q}, {_q}right{_q}{_rb}",
         'js': f"{_lb}{_q}left{_q}, {_q}down{_q}, {_q}right{_q}{_rb}"},
        # Single quotes list syntax
        {'obj': Param("[\'left\', \'down\', \'right\']", "list"),
         'py': f"{_lb}{_q}left{_q}, {_q}down{_q}, {_q}right{_q}{_rb}",
         'js': f"{_lb}{_q}left{_q}, {_q}down{_q}, {_q}right{_q}{_rb}"},
        # Single value
        {'obj': Param("\"left\"", "list"),
         'py': f"{_lb}{_q}left{_q}{_rb}",
         'js': f"{_lb}{_q}left{_q}{_rb}"},
        # Single value list syntax
        {'obj': Param("[\"left\"]", "list"),
         'py': f"{_lb}{_q}left{_q}{_rb}",
         'js': f"{_lb}{_q}left{_q}{_rb}"},
        # Variable name
        {'obj': Param("$left", "list"),
         'py': r"left",
         'js': r"left"},
    ]

    # Take note of what the script target started as
    initTarget = exputils.scriptTarget
    # Try each case
    for case in cases:
        # Test Python
        exputils.scriptTarget = "PsychoPy"
        assert (re.fullmatch(case['py'], str(case['obj'])),
                f"`{repr(case['obj'])}` should match the regex `{case['py']}`, but it was `{case['obj']}`")
        # Test JS
        exputils.scriptTarget = "PsychoJS"
        assert (re.fullmatch(case['js'], str(case['obj'])),
                f"`{repr(case['obj'])}` should match the regex `{case['js']}`, but it was `{case['obj']}`")
    # Set script target back to init
    exputils.scriptTarget = initTarget


def test_dollar_sign_syntax():
    # Define some param values, along with values which Param.dollarSyntax should return
    cases = [
        # Valid dollar and redundant dollar
        {'val': f"$hello $there",
         'ans': f"hello {_d}there",
         'valid': False},
        # Valid dollar and scaped dollar
        {'val': f"$hello \$there",
         'ans': f"hello {_sl}{_sl}{_d}there",
         'valid': False},
        # Just redundant dollar
        {'val': f"hello $there",
         'ans': f"hello {_d}there",
         'valid': False},
        # Just escaped dollar
        {'val': f"\$hello there",
         'ans': f"{_sl}{_sl}{_d}hello there",
         'valid': True},
        # Dollar in comment
        {'val': f"#$hello there",
         'ans': f"#{_d}hello there",
         'valid': False},
        # Comment after dollar
        {'val': f"$#hello there",
         'ans': f"#hello there",
         'valid': True},
        # Dollar and comment
        {'val': f"$hello #there",
         'ans': f"hello #there",
         'valid': True},
        # Valid dollar and redundtant dollar in comment
        {'val': f"$hello #$there",
         'ans': f"hello #{_d}there",
         'valid': True},
        # Valid dollar and escaped dollar in escaped d quotes
        {'val': f"$hello \"\$there\"",
         'ans': f"hello {_q}{_sl}{_sl}{_d}there{_q}",
         'valid': True},
        # Valid dollar and escaped dollar in escaped s quotes
        {'val': f"$hello \'\$there\'",
         'ans': f"hello {_q}{_sl}{_sl}{_d}there{_q}",
         'valid': True},
    ]
    # Run dollar syntax on each case
    for case in cases:
        # Make str param from val
        param = Param(case['val'], "str")
        # Run dollar syntax method
        valid, ans = param.dollarSyntax()
        # Is the output correct?
        assert re.fullmatch(case['ans'], ans), (
            f"Dollar syntax for {repr(param)} should return `{case['ans']}`, but instead returns `{ans}`"
        )
        assert valid == case['valid'], (
            f"Dollar syntax function should consider validity of {repr(param)} to be {case['valid']}, but instead "
            f"received {valid}"
        )



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
import gc
import sys

import psutil

from ..errors import print2err
from psychopy import clock

REALTIME_PRIORITY_CLASS = -18
HIGH_PRIORITY_CLASS = -10

class Computer():
    """Computer provides access to OS and Process level functionality:

    * Read the current time in sec.msec format. The time base used is shared by
      the ioHub and PsychoPy processes.
    * Access the iohub and psychopy psutil.Process objects
    * Get / set process priority and affinity.
    * Read system memory and CPU usage

    Computer contains only static methods and class attributes. Therefore
    all supported functionality can be accessed directly from the Computer
    class itself; an instance of the class never needs to be created.
    """

    #: Access to the psutil.Process class for the current system Process.
    current_process = psutil.Process()

    #: The psutil Process object for the ioHub Process.
    iohub_process = None

    #: If Computer class is on the iohub server process, psychopy_process is
    #: the psychopy process created from the pid passed to iohub on startup.
    #: The iohub server checks that this process exists
    #: (server.checkForPsychopyProcess()) and shuts down if it does not.
    psychopy_process = None

    #: The OS process ID of the ioHub Process.
    iohub_process_id = None

    #: True if the current process is the ioHub Server Process. False if the
    #: current process is the Experiment Runtime Process.
    is_iohub_process = False

    #: global_clock is used as the common time base for all devices
    #: and between the ioHub Server and Experiment Runtime Process. Do not
    #: access this class directly, instead use the Computer.getTime()
    #: and associated method name alias's to actually get the current ioHub
    # time.
    global_clock = clock.monotonicClock

    #: The name of the current operating system Python is running on.
    platform = sys.platform

    #: Python Env. bits: 32 or 64. Note that when a
    #: Python 32 bit runtime is used a 64 bit OS sysbits will equal 32.
    pybits = 32 + int(sys.maxsize > 2 ** 32) * 32

    try:
        #: Attribute representing the number of *processing units* available on
        #: the current computer. This includes cpu's, cpu cores, and hyperthreads.
        #:
        #: processing_unit_count = num_cpus * cores_per_cpu * num_hyperthreads
        #:
        #: where:
        #:      * num_cpus: Number of CPU chips on the motherboard
        #:        (usually 1 now).
        #:      * cores_per_cpu: Number of processing cores per CPU (2,4 is common)
        #:      * num_hyperthreads: Hyper-threaded cores = 2, otherwise 1.
        processing_unit_count = psutil.cpu_count()

        #: The number of cpu cores available on the computer.
        #: Hyperthreads are NOT included.
        core_count = psutil.cpu_count(False) #hyperthreads not included
    except AttributeError:
        # psutil might be too old (cpu_count added in 2.0)
        import multiprocessing
        processing_unit_count = multiprocessing.cpu_count()
        core_count = None  # but not used anyway

    try:
        #: Process priority when first started. If process is set to
        #: high priority, this value is used when the process is set back to
        #: normal priority.
        _process_original_nice_value=psutil.Process().nice() # used on linux.
    except TypeError:
        # on older versions of psutil (in ubuntu 14.04) nice is attr not call
        _process_original_nice_value=psutil.Process().nice

    #: True if the current process is currently in high or real-time
    #: priority mode (enabled by calling Computer.setPriority()).
    in_high_priority_mode = False

    def __init__(self):
        print2err('WARNING: Computer is a static class, '
                  'no need to create an instance. just use Computer.xxxxxx')

    @staticmethod
    def getPriority():
        """Returns the current processes priority as a string.

        This method is not supported on OS X.

        :return: 'normal', 'high', or 'realtime'
        """
        proc_priority = Computer.current_process.nice()
        if Computer.platform == 'win32':
            if proc_priority == psutil.HIGH_PRIORITY_CLASS:
                return 'high'
            if proc_priority == psutil.REALTIME_PRIORITY_CLASS:
                return 'realtime'
            if proc_priority == psutil.NORMAL_PRIORITY_CLASS:
                return 'normal'
        else:
            if proc_priority <= REALTIME_PRIORITY_CLASS:
                return 'realtime'
            if proc_priority <= HIGH_PRIORITY_CLASS:
                return 'high'
            if proc_priority == Computer._process_original_nice_value:
                return 'normal'
        return proc_priority

    @staticmethod
    def _setProcessPriority(process, nice_val, disable_gc):
        org_nice_val = Computer._process_original_nice_value
        try:
            process.nice(nice_val)
            Computer.in_high_priority_mode = nice_val != org_nice_val
            if disable_gc:
                gc.disable()
            else:
                gc.enable()
            return True
        except psutil.AccessDenied:
            print2err('WARNING: Could not set process {} priority '
                      'to {}'.format(process.pid, nice_val))
            return False
    @staticmethod
    def setPriority(level='normal', disable_gc=False):
        """
        Attempts to change the current processes priority based on level.
        Supported levels are:

          * 'normal': sets the current process priority to
          NORMAL_PRIORITY_CLASS on Windows, or to the processes original
          nice value on Linux.
          * 'high': sets the current process priority to HIGH_PRIORITY_CLASS
          on Windows, or to a nice value of -10 value on Linux.
          * 'realtime': sets the current process priority to
          REALTIME_PRIORITY_CLASS on Windows, or to a nice value of -18
          value on Linux.

        If level is 'normal', Python GC is also enabled.
        If level is 'high' or 'realtime', and disable_gc is True, then the
        Python garbage collection (GC) thread is suspended.

        This method is not supported on OS X.

        :return: Priority level of process when method returns.
        """
        level = level.lower()
        current_process = Computer.current_process
        nice_val = Computer._process_original_nice_value

        if level == 'normal':
            disable_gc = False
        elif level == 'high':
            nice_val = HIGH_PRIORITY_CLASS
            if Computer.platform == 'win32':
                nice_val = psutil.HIGH_PRIORITY_CLASS
        elif level.lower() == 'realtime':
            nice_val = REALTIME_PRIORITY_CLASS
            if Computer.platform == 'win32':
                nice_val = psutil.REALTIME_PRIORITY_CLASS

        Computer._setProcessPriority(current_process, nice_val, disable_gc)

        return Computer.getPriority()


    @staticmethod
    def getProcessingUnitCount():
        """
        Return the number of *processing units* available on the current
        computer.
        Processing Units include: cpu's, cpu cores, and hyper threads.

        Notes:

        * processing_unit_count = num_cpus*num_cores_per_cpu*num_hyperthreads.
        * For single core CPU's,  num_cores_per_cpu = 1.
        * For CPU's that do not support hyperthreading,  num_hyperthreads =
        1, otherwise num_hyperthreads = 2.

        Args:
            None

        Returns:
            int: the number of processing units on the computer.
        """

        return Computer.processing_unit_count

    @staticmethod
    def getProcessAffinities():
        """Retrieve the current PsychoPy Process affinity list and ioHub
        Process affinity list.

        For example, on a 2 core CPU with hyper-threading, the possible
        'processor'
        list would be [0,1,2,3], and by default both the PsychoPy and ioHub
        Processes can run on any of these 'processors', so::


            psychoCPUs,ioHubCPUS=Computer.getProcessAffinities()
            print psychoCPUs,ioHubCPUS

            >> [0,1,2,3], [0,1,2,3]


        If Computer.setProcessAffinities was used to set the PsychoPy
        Process to core 1
        (index 0 and 1) and the ioHub Process to core 2 (index 2 and 3),
        with each using both hyper threads of the given core, the set call
        would look like::


            Computer.setProcessAffinities([0,1],[2,3])

            psychoCPUs,ioHubCPUS=Computer.getProcessAffinities()
            print psychoCPUs,ioHubCPUS

            >> [0,1], [2,3]


        If the ioHub is not being used (i.e self.hub is None), then only the
        PsychoPy
        Process affinity list will be returned and None will be returned for
        the
        ioHub Process affinity::

            psychoCPUs,ioHubCPUS=Computer.getProcessAffinities()
            print psychoCPUs,ioHubCPUS

            >> [0,1,2,3], None

        **But in this case, why are you using the ioHub package at all? ;)**

        This method is not supported on OS X.

        Args:
            None

        Returns:
            (list,list) Tuple of two lists: PsychoPy Process affinity ID
            list and ioHub Process affinity ID list.

        """
        curproc_affinity = Computer.current_process.cpu_affinity()
        iohproc_affinity = Computer.iohub_process.cpu_affinity()
        return curproc_affinity, iohproc_affinity

    @staticmethod
    def setProcessAffinities(experimentProcessorList, ioHubProcessorList):
        """Sets the processor affinity for the PsychoPy Process and the ioHub
        Process.

        For example, on a 2 core CPU with hyper-threading, the possible
        'processor'
        list would be [0,1,2,3], and by default both the experiment and ioHub
        server processes can run on any of these 'processors',
        so to have both processes have all processors available
        (which is the default), you would call::

            Computer.setProcessAffinities([0,1,2,3], [0,1,2,3])

            # check the process affinities
            psychoCPUs,ioHubCPUS=Computer.getProcessAffinities()
            print psychoCPUs,ioHubCPUS

            >> [0,1,2,3], [0,1,2,3]

        based on the above CPU example.

        If setProcessAffinities was used to set the experiment process to
        core 1
        (index 0,1) and the ioHub server process to core 2 (index 2,3), with
        each using both hyper threads of the given core, the set call would
        look
        like::

            Computer.setProcessAffinities([0,1],[2,3])

            # check the process affinities
            psychoCPUs,ioHubCPUS=Computer.getProcessAffinities()
            print psychoCPUs,ioHubCPUS

            >> [0,1], [2,3]

        Args:
           experimentProcessorList (list): list of int processor ID's to set
           the PsychoPy Process affinity to. An empty list means all
           processors.

           ioHubProcessorList (list): list of int processor ID's to set the
           ioHub Process affinity to. An empty list means all processors.

        Returns:
           None

        """
        Computer.current_process.cpu_affinity(experimentProcessorList)
        Computer.iohub_process.cpu_affinity(ioHubProcessorList)

    @staticmethod
    def autoAssignAffinities():
        """Auto sets the PsychoPy Process and ioHub Process affinities based on
        some very simple logic.

        It is not known at this time if the implementation of this method makes
        any sense in terms of actually improving performance. Field tests and
        feedback will need to occur, based on which the algorithm can be
        improved.

        Currently:

        * If the system is detected to have 1 processing unit, or greater
        than 8 processing units, nothing is done by the method.
        * For a system that has two processing units, the PsychoPy Process
        is assigned to index 0, ioHub Process assigned to 1.
        * For a system that has four processing units, the PsychoPy Process
        is assigned to index's 0,1 and the ioHub Process assigned to 2,3.
        * For a system that has eight processing units, the PsychoPy Process
        is assigned to index 2,3, ioHub Process assigned to 4,5. All other
        processes running on the OS are attempted to be assigned to indexes
        0,1,6,7.

        Args:
            None

        Returns:
            None

        """
        cpu_count = Computer.processing_unit_count
        if cpu_count == 2:
            # print 'Assigning experiment process to CPU 0, ioHubServer process
            # to CPU 1'
            Computer.setProcessAffinities([0, ], [1, ])
        elif cpu_count == 4:
            # print 'Assigning experiment process to CPU 0,1, ioHubServer
            # process to CPU 2,3'
            Computer.setProcessAffinities([0, 1], [2, 3])
        elif cpu_count == 8:
            # print 'Assigning experiment process to CPU 2,3, ioHubServer
            # process to CPU 4,5, attempting to assign all others to 0,1,6,7'
            Computer.setProcessAffinities([2, 3], [4, 5])
            Computer.setAllOtherProcessesAffinity(
                    [0, 1, 6, 7],
                    [Computer.currentProcessID, Computer.iohub_process_id])
        else:
            print('autoAssignAffinities does not support %d processors.' % (
            cpu_count,))

    @staticmethod
    def getCurrentProcessAffinity():
        """
        Returns a list of 'processor' ID's (from 0 to
        Computer.processing_unit_count-1)
        that the current (calling) process is able to run on.

        Args:
            None

        Returns:
            None
        """
        return Computer.current_process.cpu_affinity()

    @staticmethod
    def setCurrentProcessAffinity(processorList):
        """
        Sets the list of 'processor' ID's (from 0 to
        Computer.processing_unit_count-1)
        that the current (calling) process should only be allowed to run on.

        Args:
           processorList (list): list of int processor ID's to set the
           current Process affinity to. An empty list means all processors.

        Returns:
            None

        """
        return Computer.current_process.cpu_affinity(processorList)

    @staticmethod
    def setProcessAffinityByID(process_id, processor_list):
        """
        Sets the list of 'processor' ID's (from 0 to
        Computer.processing_unit_count-1)
        that the process with the provided OS Process ID is able to run on.

        Args:
           processID (int): The system process ID that the affinity should
           be set for.

           processorList (list): list of int processor ID's to set process
           with the given processID too. An empty list means all processors.

        Returns:
            None
        """
        p = psutil.Process(process_id)
        return p.cpu_affinity(processor_list)

    @staticmethod
    def getProcessAffinityByID(process_id):
        """
        Returns a list of 'processor' ID's (from 0 to
        Computer.processing_unit_count-1)
        that the process with the provided processID is able to run on.

        Args:
           processID (int): The system process ID that the affinity should
           be set for.

        Returns:
           processorList (list): list of int processor ID's to set process
           with the given processID too. An empty list means all processors.
        """
        p = psutil.Process(process_id)
        return p.cpu_affinity()

    @staticmethod
    def setAllOtherProcessesAffinity(
            processor_list,
            exclude_process_id_list=[]):
        """
        Sets the affinity for all OS Processes other than those specified in
        the
        exclude_process_id_list, to the processing unit indexes specified in
        processor_list.
        Valid values in the processor_list are between 0 to
        Computer.processing_unit_count-1.

        exclude_process_id_list should be a list of OS Process ID integers,
        or an empty list (indicating to set the affiinty to all processing
        units).

        Note that the OS may not allow the calling process to set the affinity
        of every other process running on the system. For example, some system
        level processing can not have their affinity set by a user level
        application.

        However, in general, many processes can have their affinity set by
        another user process.

        Args:
           processor_list (list): list of int processor ID's to set all OS
           Processes to. An empty list means all processors.

           exclude_process_id_list (list):  A list of process ID's that
           should not have their process affinity settings changed.

        Returns:
           None
        """
        for p in psutil.pids():
            if p not in exclude_process_id_list:
                try:
                    psutil.Process(p).cpu_affinity(processor_list)
                except Exception:
                    pass

    @staticmethod
    def getProcessFromName(pnames, id_only=False):
        procs = []
        if isinstance(pnames, str):
            pnames = [pnames, ]
        for p in psutil.process_iter():
            if p.name() in pnames:
                if id_only:
                    procs.append(p.pid)
                else:
                    procs.append(p)
        return procs

    @staticmethod
    def getTime():
        """
        Returns the current sec.msec-msec time of the system.

        The underlying timer that is used is based on OS and Python version.
        Three requirements exist for the ioHub time base implementation:

        * The Python interpreter does not apply an offset to the times
        returned based on when the timer module being used was loaded or
        when the timer function first called was first called.
        * The timer implementation used must be monotonic and report elapsed
        time between calls, 'not' CPU usage time.
        * The timer implementation must provide a resolution of 50 usec or
        better.

        Given the above requirements, ioHub selects a timer implementation
        as follows:

        * On Windows, the Windows Query Performance Counter API is used
        using ctypes access.
        * On other OS's, if the Python version being used is 2.6 or lower,
        time.time is used. For Python 2.7 and above,
        the timeit.default_timer function is used.

        Args:
           None

        Returns:
           None
        """
        return Computer.global_clock.getTime()

    @staticmethod
    def getPhysicalSystemMemoryInfo():
        """Return a class containing information about current memory usage.

        Args:
           None

        Returns:
           vmem: (total=long, available=long, percent=float, used=long,
           free=long)

        Where:

        * vmem.total: the total amount of memory in bytes.
        * vmem.available: the available amount of memory in bytes.
        * vmem.percent: the percent of memory in use by the system.
        * vmem.used: the used amount of memory in bytes.
        * vmem.free: the amount of memory that is free in bytes.On Windows,
        this is the same as vmem.available.

        """
        m = psutil.virtual_memory()
        return m

    @staticmethod
    def getCPUTimeInfo(percpu=False):
        """Return a float representing the current CPU utilization as a
        percentage.

        Args:
           percpu (bool): If True, a list of cputimes objects is returned,
                          one for each processing unit for the computer.
                          If False, only a single cputimes object is returned.
        Returns:
           object: (user=float, system=float, idle=float)

        """
        return psutil.cpu_times_percent(percpu=percpu)

    @staticmethod
    def getCurrentProcess():
        """Get the current / Local process.

        On Windows and Linux, this is a psutil.Process class instance.

        Args:
           None

        Returns:
           object: Process object for the current system process.

        """
        return Computer.current_process

    @staticmethod
    def getIoHubProcess():
        """Get the ioHub Process.

        On Windows and Linux, this is a psutil.Process class instance.

        Args:
           None

        Returns:
           object: Process object for the ioHub Process.

        """
        return Computer.iohub_process



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tools for working with audio data.

This module provides routines for saving/loading and manipulating audio samples.

"""

__all__ = [
    'array2wav',
    'wav2array',
    'sinetone',
    'squaretone',
    'sawtone',
    'whiteNoise',
    'audioBufferSize',
    'SAMPLE_RATE_8kHz', 'SAMPLE_RATE_TELCOM_QUALITY',
    'SAMPLE_RATE_16kHz', 'SAMPLE_RATE_VOIP_QUALITY', 'SAMPLE_RATE_VOICE_QUALITY',
    'SAMPLE_RATE_22p05kHz', 'SAMPLE_RATE_AM_RADIO_QUALITY',
    'SAMPLE_RATE_32kHz', 'SAMPLE_RATE_FM_RADIO_QUALITY',
    'SAMPLE_RATE_44p1kHz', 'SAMPLE_RATE_CD_QUALITY',
    'SAMPLE_RATE_48kHz', 'SAMPLE_RATE_DVD_QUALITY',
    'SAMPLE_RATE_96kHz',
    'SAMPLE_RATE_192kHz'
]

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import os
import numpy as np
from scipy.io import wavfile
from scipy import signal

# pydub is needed for saving and loading MP3 files among others
# _has_pydub = True
# try:
#     import pydub
# except (ImportError, ModuleNotFoundError):
#     _has_pydub = False

# Constants for common sample rates. Some are aliased to give the programmer an
# idea to the quality they would expect from each. It is recommended to only use
# these values since most hardware supports them for recording and playback.
#
SAMPLE_RATE_8kHz = SAMPLE_RATE_TELCOM_QUALITY = 8000
SAMPLE_RATE_16kHz = SAMPLE_RATE_VOIP_QUALITY = SAMPLE_RATE_VOICE_QUALITY = 16000
SAMPLE_RATE_22p05kHz = SAMPLE_RATE_AM_RADIO_QUALITY = 22050
SAMPLE_RATE_32kHz = SAMPLE_RATE_FM_RADIO_QUALITY = 32000  # wireless headphones
SAMPLE_RATE_44p1kHz = SAMPLE_RATE_CD_QUALITY = 44100
SAMPLE_RATE_48kHz = SAMPLE_RATE_DVD_QUALITY = 48000
SAMPLE_RATE_96kHz = 96000
SAMPLE_RATE_192kHz = 192000  # high-def

# needed for converting float to int16, not exported by __all__
MAX_16BITS_SIGNED = 1 << 15


def array2wav(filename, samples, freq=48000):
    """Write audio samples stored in an array to WAV file.

    Parameters
    ----------
    filename : str
        File name for the output.
    samples : ArrayLike
        Nx1 or Nx2 array of audio samples with values ranging between -1 and 1.
    freq : int or float
        Sampling frequency used to capture the audio samples in Hertz (Hz).
        Default is 48kHz (specified as `48000`) which is considered DVD quality
        audio.

    """
    # rescale
    clipData = np.asarray(samples * (MAX_16BITS_SIGNED - 1), dtype=np.int16)

    # write out file
    wavfile.write(filename, freq, clipData)


def wav2array(filename, normalize=True):
    """Read a WAV file and write samples to an array.

    Parameters
    ----------
    filename : str
        File name for WAV file to read.
    normalize : bool
        Convert samples to floating point format with values ranging between
        -1 and 1. If `False`, values will be kept in `int16` format. Default is
        `True` since normalized floating-point is the typical format for audio
        samples in PsychoPy.

    Returns
    -------
    samples : ArrayLike
        Nx1 or Nx2 array of samples.
    freq : int
        Sampling frequency for playback specified by the audio file.

    """
    fullpath = os.path.abspath(filename)  # get the full path

    if not os.path.isfile(fullpath):  # check if the file exists
        raise FileNotFoundError(
            "Cannot find WAV file `{}` to open.".format(filename))

    # read the file
    freq, samples = wavfile.read(filename, mmap=False)

    # transpose samples
    samples = samples[:, np.newaxis]

    # check if we need to normalize things
    if normalize:
        samples = np.asarray(
            samples / (MAX_16BITS_SIGNED - 1), dtype=np.float32)

    return samples, int(freq)


def sinetone(duration, freqHz, gain=0.8, sampleRateHz=SAMPLE_RATE_48kHz):
    """Generate audio samples for a tone with a sine waveform.

    Parameters
    ----------
    duration : float or int
        Length of the sound in seconds.
    freqHz : float or int
        Frequency of the tone in Hertz (Hz). Note that this differs from the
        `sampleRateHz`.
    gain : float
        Gain factor ranging between 0.0 and 1.0.
    sampleRateHz : int
        Samples rate of the audio for playback.

    Returns
    -------
    ndarray
        Nx1 array containing samples for the tone (single channel).

    """
    assert 0.0 <= gain <= 1.0   # check if gain range is valid

    nsamp = sampleRateHz * duration
    samples = np.arange(nsamp, dtype=np.float32)
    samples[:] = 2 * np.pi * samples[:] * freqHz / sampleRateHz
    samples[:] = np.sin(samples)

    if gain != 1.0:
        samples *= gain

    return samples.reshape(-1, 1)


def squaretone(duration, freqHz, dutyCycle=0.5, gain=0.8,
               sampleRateHz=SAMPLE_RATE_48kHz):
    """Generate audio samples for a tone with a square waveform.

    Parameters
    ----------
    duration : float or int
        Length of the sound in seconds.
    freqHz : float or int
        Frequency of the tone in Hertz (Hz). Note that this differs from the
        `sampleRateHz`.
    dutyCycle : float
        Duty cycle between 0.0 and 1.0.
    gain : float
        Gain factor ranging between 0.0 and 1.0.
    sampleRateHz : int
        Samples rate of the audio for playback.

    Returns
    -------
    ndarray
        Nx1 array containing samples for the tone (single channel).

    """
    assert 0.0 <= gain <= 1.0  # check if gain range is valid

    nsamp = sampleRateHz * duration
    samples = np.arange(nsamp, dtype=np.float32)
    samples[:] = 2 * np.pi * samples[:] * freqHz / sampleRateHz
    samples[:] = signal.square(samples, duty=dutyCycle)

    if gain != 1.0:
        samples *= gain

    return samples.reshape(-1, 1)


def sawtone(duration, freqHz, peak=0.5, gain=0.8,
            sampleRateHz=SAMPLE_RATE_48kHz):
    """Generate audio samples for a tone with a sawtooth waveform.

    Parameters
    ----------
    duration : float or int
        Length of the sound in seconds.
    freqHz : float or int
        Frequency of the tone in Hertz (Hz). Note that this differs from the
        `sampleRateHz`.
    peak : float
        Location of the peak between 0.0 and 1.0. If the peak is at 0.5, the
        resulting wave will be triangular. A value of 1.0 will cause the peak to
        be located at the very end of a cycle.
    gain : float
        Gain factor ranging between 0.0 and 1.0.
    sampleRateHz : int
        Samples rate of the audio for playback.

    Returns
    -------
    ndarray
        Nx1 array containing samples for the tone (single channel).

    """
    assert 0.0 <= gain <= 1.0  # check if gain range is valid

    nsamp = sampleRateHz * duration
    samples = np.arange(nsamp, dtype=np.float32)
    samples[:] = 2 * np.pi * samples[:] * freqHz / sampleRateHz
    samples[:] = signal.sawtooth(samples, width=peak)

    if gain != 1.0:
        samples *= gain

    return samples.reshape(-1, 1)


def whiteNoise(duration=1.0, sampleRateHz=SAMPLE_RATE_48kHz):
    """Generate gaussian white noise.

    Parameters
    ----------
    duration : float or int
        Length of the sound in seconds.
    sampleRateHz : int
        Samples rate of the audio for playback.

    Returns
    -------
    ndarray
        Nx1 array containing samples for the sound.

    """
    samples = np.random.randn(int(duration * sampleRateHz)).reshape(-1, 1)

    # clip range
    samples = samples.clip(-1, 1)

    return samples


def audioBufferSize(duration=1.0, freq=SAMPLE_RATE_48kHz):
    """Estimate the memory footprint of an audio clip of given duration. Assumes
    that data is stored in 32-bit floating point format.

    This can be used to determine how large of a buffer is needed to store
    enough samples for `durations` seconds of audio using the specified
    frequency (`freq`).

    Parameters
    ----------
    duration : float
        Length of the clip in seconds.
    freq : int
        Sampling frequency in Hz.

    Returns
    -------
    int
        Estimated number of bytes.

    """
    # Right now we are just computing for single precision floats, we can expand
    # this to other types in the future.

    sizef32 = 32  # duh

    return int(duration * freq * sizef32)


if __name__ == "__main__":
    pass



"""
This demo produces a handy diagram showing how the metrics of a PsychoPy GLFont object affect the layout of rows in a textbox. A textbox is drawn such that the baseline of the first line is at the vertical coordinate 0, meaning that all other lines and shapes can be laid out relative to this line.
"""


from psychopy import visual, event
from psychopy.visual.textbox2.fontmanager import FontManager

# Create window
win = visual.Window(size=(500, 200), units="pix", color="white")

# Create a list to store objects for easy drawing
drawList = []

# Get a font with consistent proportions that are easy to spot
allFonts = FontManager()
font = allFonts.getFont("Outfit", size=50, lineSpacing=1.3)

# Create a textbox using this font, whose vertical position is such that the baseline of the first line of text is at 0
text = visual.TextBox2(
    win=win, text="My text has an È!\nMy text has an È!", font=font,
    pos=(-50, font.ascender), size=(400, font.height), padding=0,
    color="black", anchor="top center", alignment="top left"
)
drawList += [text]

# Draw the baseline
baseline = visual.Line(win, start=(-250, 0), end=(250, 0), lineColor="green")
baselineLbl = visual.TextBox2(win, "baseline (0)", "Outfit", color="green", letterHeight=12, pos=(160, 8), size=(50, 10), padding=0)
drawList += [baseline, baselineLbl]
# Draw the descent line
descender = visual.Line(win, start=(-250, font.descender), end=(250, font.descender), lineColor="blue")
descenderLbl = visual.TextBox2(win, ".descender", "Outfit", color="blue", letterHeight=12, pos=(160, font.descender + 8), size=(50, 10), padding=0)
drawList += [descender, descenderLbl]
# Draw the ascent line
ascender = visual.Line(win, start=(-250, font.ascender), end=(250, font.ascender), lineColor="orange")
ascenderLbl = visual.TextBox2(win, ".ascender", "Outfit", color="orange", letterHeight=12, pos=(160, font.ascender + 8), size=(50, 10), padding=0)
drawList += [ascender, ascenderLbl]
# Draw the cap height line
capheight = visual.Line(win, start=(-250, font.capheight), end=(250, font.capheight), lineColor="red")
capheightLbl = visual.TextBox2(win, ".capheight", "Outfit", color="red", letterHeight=12, pos=(160, font.capheight + 8), size=(50, 10), padding=0)
drawList += [capheight, capheightLbl]
# Draw the leading line
leading = visual.Line(win, start=(-250, font.leading), end=(250, font.leading), lineColor="purple")
leadingLbl = visual.TextBox2(win, ".leading", "Outfit", color="purple", letterHeight=12, pos=(160, font.leading - 4), size=(50, 10), padding=0)
drawList += [leading, leadingLbl]

# Draw the height box
height = visual.Rect(win, fillColor="orange", pos=(215, (font.ascender + font.leading)/2), size=(20, font.height))
heightLbl = visual.TextStim(win, ".height", "Outfit", color="white", bold=True, pos=(215, (font.ascender + font.leading)/2), height=12, ori=90)
drawList += [height, heightLbl]
# Draw the size box
size = visual.Rect(win, fillColor="red", pos=(240, (font.capheight + font.descender)/2), size=(20, font.size))
sizeLbl = visual.TextStim(win, ".size", "Outfit", color="white", bold=True, pos=(240, (font.capheight + font.descender)/2), height=12, ori=90)
drawList += [size, sizeLbl]
# Draw linegap box
linegap = visual.TextBox2(win, ".linegap", "Outfit", fillColor="purple", color="white", letterHeight=12, pos=(160, (font.descender + font.leading)/2), size=(45, font.linegap), padding=0)
drawList += [linegap]

# Rearrange order
drawList += [drawList.pop(drawList.index(descender))]
drawList += [drawList.pop(drawList.index(height))]
drawList += [drawList.pop(drawList.index(heightLbl))]
drawList += [drawList.pop(drawList.index(leading))]

while not event.getKeys("escape"):
    for obj in drawList:
        obj.draw()
    win.flip()
